# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Draws detected groups and their tags
"""

from typing import Any, Dict, List

import numpy as np
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.draw.utils.bbox import draw_bboxes, draw_tags
from peekingduck.pipeline.nodes.draw.utils.constants import TOMATO


class Node(AbstractNode):
    """This node draws large bounding boxes over multiple object bounding boxes
    which have been identified as belonging to large groups.

    The draw group_bbox_and_tage node uses the obj_groups and large_groups from the
    dabble nodes to draw group bboxes and the large group message tag onto the image.
    For better understanding, refer to
    `group size checking usecase <use_cases/group_size_checking.html>`_.

    Inputs:

        |img|

        |bboxes|

        |obj_groups|

        |large_groups|

    Outputs:
        |none|

    Configs:
        tag (:obj:`str`): **default = "LARGE GROUP!"**
            string message printed in the case of a large group detected.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """ Draws large bounding boxes over multiple object bounding boxes
        which have been identified as belonging to large groups

        Args:
            inputs (dict): Dict with keys "img", "bboxes", "obj_groups",
                "large_groups".

        Returns:
            outputs (dict): Dict with keys "none".
        """

        group_bboxes = self._get_group_bbox_coords(
            inputs["large_groups"], inputs["bboxes"], inputs["obj_groups"])
        group_tags = self._get_group_tags(
            inputs["large_groups"], self.tag)

        # show labels set to False to reduce clutter on display
        draw_bboxes(inputs["img"], group_bboxes, TOMATO, False)   # type: ignore
        draw_tags(inputs["img"], group_bboxes, group_tags, TOMATO)  # type: ignore

        return {}

    @staticmethod
    def _get_group_bbox_coords(large_groups: List[int],
                               bboxes: List[np.array],
                               obj_groups: List[int]) -> List[np.array]:
        """ For bboxes that belong to the same large group, get the coordinates of
        a large bbox that combines all these individual bboxes. Repeat for all large
        groups.
        """
        group_bboxes = []
        bboxes = np.array(bboxes)
        obj_groups = np.array(obj_groups)
        for group in large_groups:
            # filter relevant bboxes, select top-left and bot-right corners
            group_bbox = np.array([1., 1., 0., 0.])
            selected_bboxes = bboxes[obj_groups == group]
            group_bbox[:2] = np.amin(selected_bboxes, axis=0)[:2]
            group_bbox[2:] = np.amax(selected_bboxes, axis=0)[2:]
            group_bboxes.append(group_bbox)

        return group_bboxes

    @staticmethod
    def _get_group_tags(large_groups: List[int], tag: str) -> List[str]:
        """ Creates a list of tags to be used for the draw_tags function.
        """

        group_tags = [tag for _ in range(len(large_groups))]

        return group_tags



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EfficientDet model with model types: D0-D4
"""

import logging
from typing import Dict, Any, List, Tuple
import numpy as np

from peekingduck.weights_utils import checker, downloader
from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files.detector import Detector


class EfficientDetModel:
    """EfficientDet model with model types: D0-D4"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__()

        self.logger = logging.getLogger(__name__)

        # check threshold values
        if not 0 <= config['score_threshold'] <= 1:
            raise ValueError("score_threshold must be in [0, 1]")
        if not 0 <= config['model_type'] <= 4:
            raise ValueError("model_type must be an integer in [0, 4]")

        # check for efficientdet weights, if none then download into weights folder
        if not checker.has_weights(config['root'],
                                   config['weights_dir']):
            self.logger.info('---no efficientdet weights detected. proceeding to download...---')
            downloader.download_weights(config['root'],
                                        config['blob_file'])
            self.logger.info('---efficientdet weights download complete.---')

        self.detect_ids = config['detect_ids']
        self.logger.info('efficientdet model detecting ids: %s', self.detect_ids)

        self.detector = Detector(config)

    def predict(self, frame: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """predict the bbox from frame

        returns:
        object_bboxes(np.ndarray): list of bboxes detected
        object_labels(np.ndarray): list of index labels of the
            object detected for the corresponding bbox
        object_scores(np.ndarray): list of confidence scores of the
            object detected for the corresponding bbox
        """
        assert isinstance(frame, np.ndarray)

        # return bboxes, object_bboxes, object_labels, object_scores
        return self.detector.predict_bbox_from_image(frame, self.detect_ids)

    def get_detect_ids(self) -> List[int]:
        """getter function for ids to be detected
        """
        return self.detect_ids



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Functions to download model weights
"""

import os
import zipfile
import requests
from tqdm import tqdm


BASE_URL = "https://peekingduck.blob.core.windows.net/models"


def download_weights(root: str, blob_file: str) -> None:
    """Download weights for specified blob_file

    Args:
        root (str): root directory of peekingduck
        blob_file (str): name of file to be downloaded
    """

    extract_dir = os.path.join(root, "..", "weights")
    zip_path = os.path.join(root, "..", "weights", "temp.zip")

    download_file_from_blob(blob_file, zip_path)

    # search for downloaded .zip file and extract, then delete
    with zipfile.ZipFile(zip_path, "r") as temp:
        for file in tqdm(iterable=temp.namelist(), total=len(temp.namelist())):
            temp.extract(member=file, path=extract_dir)

    os.remove(zip_path)


def download_file_from_blob(file_name: str, destination: str) -> None:
    """Method to download publicly shared files from azure blob

    Args:
        file_name (str): name of file to be downloaded
        destination (str): destination directory of download
    """

    url = f"{BASE_URL}/{file_name}"
    session = requests.Session()

    response = session.get(url, stream=True)
    save_response_content(response, destination)


def save_response_content(response: requests.Response, destination: str) -> None:
    """Chunk saving of download content. Chunk size set to large
    integer as weights are usually pretty large

    Args:
        response (Reponse): html response
        destination (str): destintation directory of download
    """
    chunk_size = 32768

    with open(destination, "wb") as temp:
        for chunk in tqdm(response.iter_content(chunk_size)):
            if chunk:  # filter out keep-alive new chunks
                temp.write(chunk)



# Copyright 2021 AI Singapore

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Write the output image/video to file
"""

import datetime
import os
from typing import Any, Dict
import numpy as np
import cv2
from peekingduck.pipeline.nodes.node import AbstractNode

# role of this node is to be able to take in multiple frames, stitch them together and output them.
# to do: need to have 'live' kind of data when there is no filename
# to do: it will be good to have the accepted file format as a configuration
# to do: somewhere so that input and output can use this config for media related issues


class Node(AbstractNode):
    """Node that outputs the processed image or video to a file.

    Inputs:
        |img|

        |filename|

        |saved_video_fps|

        |pipeline_end|

    Outputs:
        None

    Configs:
        output_dir (:obj:`str`): **default = 'PeekingDuck/data/output'**

            Output directory for files to be written locally.
    """

    def __init__(self, config: Dict[str, Any]=None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

        self._file_name = None
        self._prepare_directory(self.output_dir)
        self._fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        self._image_type = None
        self._file_path = None
        self.writer = None
        self._file_path_with_timestamp = None
        self.logger.info('Output directory used is: %s', self.output_dir)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """ Writes media information to filepath"""

        # reset and terminate when there are no more data
        if inputs["pipeline_end"]:
            if self.writer:  # images automatically releases writer
                self.writer.release()
            return {}

        if not self._file_name:
            self._prepare_writer(inputs["filename"],
                                 inputs["img"],
                                 inputs["saved_video_fps"])

        if inputs["filename"] != self._file_name:
            self._prepare_writer(inputs["filename"],
                                 inputs["img"],
                                 inputs["saved_video_fps"])

        self._write(inputs["img"])

        return {}

    def _write(self, img: np.array) -> None:
        if self._image_type == "image":
            cv2.imwrite(self._file_path_with_timestamp, img)
        else:
            self.writer.write(img)  # type: ignore

    def _prepare_writer(self, filename: str, img: np.array, saved_video_fps: int) -> None:

        self._file_path_with_timestamp = self._append_datetime_filename(filename)  # type: ignore

        self._image_type = "video"  # type: ignore
        if filename.split(".")[-1] in ["jpg", "jpeg", "png"]:
            self._image_type = "image"  # type: ignore
        else:
            resolution = img.shape[1], img.shape[0]
            self.writer = cv2.VideoWriter(
                self._file_path_with_timestamp, self._fourcc, saved_video_fps, resolution)

    @staticmethod
    def _prepare_directory(output_dir) -> None:  # type: ignore
        os.makedirs(output_dir, exist_ok=True)

    def _append_datetime_filename(self, filename: str) -> str:

        self._file_name = filename  # type: ignore
        current_time = datetime.datetime.now()
        time_str = current_time.strftime("%d%m%y-%H-%M-%S")  # output as '240621-15-09-13'

        # append timestamp to filename before extension Format: filename_timestamp.extension
        filename_with_timestamp = filename.split(".")[-2] \
            + "_" + time_str \
            + "."+filename.split(".")[-1]
        file_path_with_timestamp = os.path.join(  # type: ignore
            self.output_dir, filename_with_timestamp)

        return file_path_with_timestamp



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Main class for HRNet Model
"""


import logging
from typing import Any, Dict, Tuple

import numpy as np
from peekingduck.weights_utils import checker, downloader
from peekingduck.pipeline.nodes.model.hrnetv1.hrnet_files.detector import Detector


class HRNetModel:  # pylint: disable=too-few-public-methods
    """HRNet model to detect poses from detected bboxes
    """

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__()

        self.logger = logging.getLogger(__name__)

        # check threshold values
        if not 0 <= config['score_threshold'] <= 1:
            raise ValueError("score_threshold must be in [0, 1]")

        # check for hrnet weights, if none then download into weights folder
        if not checker.has_weights(config['root'], config['weights_dir']):
            print('---no hrnet weights detected. proceeding to download...---')
            downloader.download_weights(config['root'], config['blob_file'])
            print('---hrnet weights download complete.---')

        self.detector = Detector(config)
        self.threshold_score = config['score_threshold']

    def predict(self, frame: np.ndarray, bboxes: np.ndarray) -> \
            Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Predict poses from input frame and bboxes

        Args:
            frame (np.array): image in numpy array
            bboxes (np.array): detected bboxes in image

        Returns:
            keypoints, keypoint_scores, keypoint_conns, keypoint_bboxes
            (Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]): \
            tuple containing list of bboxes and pose related info i.e coordinates,
            scores, connections
        """
        assert isinstance(frame, np.ndarray)
        assert isinstance(bboxes, np.ndarray)
        detected_bboxes = bboxes.copy()
        if bboxes.size != 0:
            return self.detector.predict(frame, detected_bboxes)
        keypoints = np.array([])
        keypoint_scores = np.array([])
        keypoint_conns = np.array([])

        return keypoints, keypoint_scores, keypoint_conns



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Draws the 2D boundaries of a zone
"""

from typing import Any, Dict
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.draw.utils.zone import draw_zones


class Node(AbstractNode):
    """Draws the boundaries of each specified zone onto the image.

    The draw zones node uses the zones from the zone_count dabble node to
    draw a bounding box that represents the zone boundaries onto the image.

    Inputs:

        |img|

        |zones|

    Outputs:
        |none|

    Configs:
        None.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Draws the boundaries of each specified zone onto the image.

        Args:
            inputs (dict): Dict with keys "zones", "img".

        Returns:
            outputs (dict): Dict with keys "none".
        """

        draw_zones(inputs["img"], inputs["zones"])  # type: ignore
        return {}



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Calculates the FPS of video
"""

from typing import Any, Dict, List
from statistics import mean
from time import perf_counter

from peekingduck.pipeline.nodes.node import AbstractNode

NUM_FRAMES = 10


class Node(AbstractNode):
    """ FPS node class that calculates the FPS of the image frame.

    This node calculates instantaneous FPS and a 10 frame moving average
    FPS. A preferred output setting can be set via the config file.

    Inputs:
        |pipeline_end|

    Outputs:
        |fps|

    Configs:
        fps_log_display (:obj:`bool`): **default="False"**

            Enables logging of 10 frame moving average FPS during execution of peekingduck

        fps_log_freq (:obj:`int`): **default="100"**
            Frequency of logging moving average FPS every n frames

        dampen_fps (:obj:`bool`): **default="True"**
            Returns moving average FPS if True, instantaneous FPS if False

    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

        self.count = 0
        self.global_avg_fps = 0.0
        self.prev_frame_timestamp = 0.0
        self.moving_average_fps: List[float] = []

        if self.fps_log_display:
            self.logger.info('Moving average of FPS will be logged every: %s frames',
                             self.fps_log_freq)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """ Calculates FPS using the time difference between the current
        frame and the previous frame."""

        curr_frame_timestamp = perf_counter()

        # Frame level FPS
        frame_fps = 1./(curr_frame_timestamp - self.prev_frame_timestamp)
        self.prev_frame_timestamp = curr_frame_timestamp

        # Calculate moving average FPS (dampen_fps)
        average_fps = self._moving_average(frame_fps)

        # Ignore FPS of final frame to avoid skewing final average
        if not inputs["pipeline_end"]:
            if self.fps_log_display:
                if self.count % self.fps_log_freq == 0 and self.count != 0:
                    self.logger.info(
                        'Avg FPS over last 10 frames: %.2f', average_fps)

            # Calculate FPS over all processed frames
            self.global_avg_fps = self._global_average(frame_fps)

        if inputs["pipeline_end"]:
            self.logger.info('Avg FPS over all processed frames: %.2f',
                             self.global_avg_fps)

        return {"fps": average_fps} if self.dampen_fps else {"fps": frame_fps}

    def _moving_average(self, frame_fps: float) -> float:
        self.moving_average_fps.append(frame_fps)
        if len(self.moving_average_fps) > NUM_FRAMES:
            self.moving_average_fps.pop(0)
        moving_average_val = mean(self.moving_average_fps)
        return moving_average_val

    def _global_average(self, frame_fps: float) -> float:
        # Cumulative moving average formula
        global_average = \
            (frame_fps + self.count*self.global_avg_fps) / (self.count + 1)
        self.count += 1
        return global_average



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Check if number of objects in a group exceed a threshold
"""

from typing import Any, Dict, List
from collections import Counter

from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """This node checks which groups have exceeded the group size threshold.

    Inputs:
        |obj_groups|

    Outputs:
        |large_groups|

    Configs:
        group_size_thres (:obj:`int`): **default = 5**

            Threshold of group size.

    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, List[int]]) -> Dict[str, List[int]]:
        """ Checks which groups have exceeded the group size threshold,
        and returns a list of such groups.
        """

        group_counter = Counter(inputs["obj_groups"])
        large_groups = [group for group in group_counter if
                        group_counter[group] > self.group_size_thres]

        return {"large_groups": large_groups}



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Show the outputs on your display
"""

from typing import Any, Dict
import cv2
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Node that streams the output on your display

    Inputs:
        |img|

    Outputs:
        |pipeline_end|

    """

    def __init__(self, config: Dict[str, Any]=None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """ Show the outputs on your display"""

        cv2.imshow('PeekingDuck', inputs["img"])

        if cv2.waitKey(1) & 0xFF == ord('q'):
            return {"pipeline_end": True}

        return {"pipeline_end": False}



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Constants used for PoseNet
"""

import numpy as np

KEYPOINT_NAMES = (
    "nose", "leftEye", "rightEye", "leftEar", "rightEar", "leftShoulder",
    "rightShoulder", "leftElbow", "rightElbow", "leftWrist", "rightWrist",
    "leftHip", "rightHip", "leftKnee", "rightKnee", "leftAnkle", "rightAnkle"
)

_KEYPOINT_IDS = {pn: pid for pid, pn in enumerate(KEYPOINT_NAMES)}

_POSE_CHAIN = [
    ("nose", "leftEye"), ("leftEye", "leftEar"), ("nose", "rightEye"),
    ("rightEye", "rightEar"), ("nose", "leftShoulder"),
    ("leftShoulder", "leftElbow"), ("leftElbow", "leftWrist"),
    ("leftShoulder", "leftHip"), ("leftHip", "leftKnee"),
    ("leftKnee", "leftAnkle"), ("nose", "rightShoulder"),
    ("rightShoulder", "rightElbow"), ("rightElbow", "rightWrist"),
    ("rightShoulder", "rightHip"), ("rightHip", "rightKnee"),
    ("rightKnee", "rightAnkle")
]

POSE_CONNECTIONS = np.array([(_KEYPOINT_IDS[parent], _KEYPOINT_IDS[child])
                             for parent, child in _POSE_CHAIN])

SKELETON = [[16, 14], [14, 12], [17, 15], [15, 13], [12, 13],
            [6, 12], [7, 13], [6, 7], [6, 8], [7, 9],
            [8, 10], [9, 11], [2, 3], [1, 2], [1, 3], [2, 4],
            [3, 5], [4, 6], [5, 7]]

LOCAL_MAXIMUM_RADIUS = 1

SCALE_FACTOR = 1.0

KEYPOINTS_NUM = 17

MIN_PART_SCORE = 0.1

IMAGE_NET_MEAN = [-123.15, -115.90, -103.06]

SWAP_AXES = (1, 0)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Reader functions for input nodes
"""

from typing import Any, Tuple, Union
from threading import Thread, Lock
import cv2
from peekingduck.pipeline.nodes.input.utils.preprocess import mirror


class VideoThread:
    '''
    Videos will be threaded to prevent I/O blocking from affecting FPS.
    '''

    def __init__(self, input_source: str, mirror_image: bool) -> None:
        self.stream = cv2.VideoCapture(input_source)
        self.mirror = mirror_image
        if not self.stream.isOpened():
            raise ValueError("Camera or video input not detected: %s" % input_source)

        self._lock = Lock()

        self.frame = None
        thread = Thread(target=self._reading_thread, args=(), daemon=True)
        thread.start()

    def __del__(self) -> None:
        self.stream.release()

    def _reading_thread(self) -> None:
        '''
        A thread that continuously polls the camera for frames.
        '''
        while True:
            _, self.frame = self.stream.read()

    def read_frame(self) -> Union[bool, Any]:
        '''
        Reads the frame.
        '''
        self._lock.acquire()
        if self.frame is not None:
            frame = self.frame.copy()
            self._lock.release()
            if self.mirror:
                frame = mirror(frame)
            return True, frame

        self._lock.release()
        return False, None

    @property
    def resolution(self) -> Tuple[int, int]:
        """ Get resolution of the camera device used.

        Returns:
            width(int): width of input resolution
            height(int): heigh of input resolution
        """
        width = self.stream.get(cv2.CAP_PROP_FRAME_WIDTH)
        height = self.stream.get(cv2.CAP_PROP_FRAME_HEIGHT)
        return int(width), int(height)


class VideoNoThread:
    '''
    No threading to deal with recorded videos and images.
    '''

    def __init__(self, input_source: str, mirror_image: bool) -> None:
        self.stream = cv2.VideoCapture(input_source)
        self.mirror = mirror_image
        if not self.stream.isOpened():
            raise ValueError("Video or image path incorrect: %s" % input_source)

    def __del__(self) -> None:
        self.stream.release()

    def read_frame(self) -> None:
        '''
        Reads the frame.
        '''
        return self.stream.read()

    @property
    def fps(self) -> float:
        """ Get FPS of videofile

        Returns:
            int: number indicating FPS
        """
        fps = self.stream.get(cv2.CAP_PROP_FPS)
        return fps

    @property
    def resolution(self) -> Tuple[int, int]:
        """ Get resolution of the file.

        Returns:
            width(int): width of resolution
            height(int): heigh of resolution
        """
        width = self.stream.get(cv2.CAP_PROP_FRAME_WIDTH)
        height = self.stream.get(cv2.CAP_PROP_FRAME_HEIGHT)
        return int(width), int(height)

    @property
    def frame_count(self) -> int:
        """ Get total number of frames of file

        Returns:
            int: number indicating frame count
        """
        num_frames = self.stream.get(cv2.CAP_PROP_FRAME_COUNT)
        return int(num_frames)



# Copyright 2018 Ross Wightman
# Modifications copyright 2021 AI Singapore
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      https://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Support functions to decode single pose
"""


from typing import Tuple
import numpy as np

from peekingduck.pipeline.nodes.model.posenetv1.posenet_files.constants import POSE_CONNECTIONS


def decode_pose(root_score: float,
                root_id: int,
                root_image_coord: np.ndarray,
                scores: np.ndarray,
                offsets: np.ndarray,
                output_stride: int,
                displacements_fwd: np.ndarray,
                displacements_bwd: np.ndarray,
                keypoint_scores: np.ndarray,
                keypoint_coords: np.ndarray) -> None:
    # pylint: disable=too-many-arguments
    """ Decode pose's keypoints scores and coordinates from keypoints score,
    coordinates and displacements
    Args:
        root_score (float): a keypoint with highest score is selected as root
        root_id (int): root keypoint's index
        root_image_coord (np.array): relative coordinate of root keypoint
        scores (np.array): HxWxNP heatmap scores of NP body parts
        offsets (np.array): HxWxNPx2 short range offset vector of NP body parts
        output_stride (int): output stride to convert output indices to image coordinates
        displacements_fwd (np.array): HxWxNEx2 forward displacements of NE body
                connections
        displacements_bwd (np.array): HxWxNEx2 backward displacements of NE body
                connections
        keypoints_scores (np.array): 17x1 buffer to store keypoint scores
        keypoint_coords (np.array): 17x2 buffer to store keypoint coordinates
    Returns:
        pose_count (int): number of poses detected
    """
    num_edges = len(POSE_CONNECTIONS)

    keypoint_scores[root_id] = root_score
    keypoint_coords[root_id] = root_image_coord

    for edge in reversed(range(num_edges)):
        target_keypoint_id, source_keypoint_id = POSE_CONNECTIONS[edge]
        _calculate_instance_keypoints(edge,
                                      target_keypoint_id,
                                      source_keypoint_id,
                                      keypoint_scores,
                                      keypoint_coords,
                                      scores,
                                      offsets,
                                      output_stride,
                                      displacements_bwd)

    for edge in range(num_edges):
        source_keypoint_id, target_keypoint_id = POSE_CONNECTIONS[edge]
        _calculate_instance_keypoints(edge,
                                      target_keypoint_id,
                                      source_keypoint_id,
                                      keypoint_scores,
                                      keypoint_coords,
                                      scores,
                                      offsets,
                                      output_stride,
                                      displacements_fwd)


def _calculate_instance_keypoints(edge: int,
                                  target_keypoint_id: int,
                                  source_keypoint_id: int,
                                  instance_keypoint_scores: np.ndarray,
                                  instance_keypoint_coords: np.ndarray,
                                  scores: np.ndarray,
                                  offsets: np.ndarray,
                                  output_stride: int,
                                  displacements: np.ndarray) -> None:
    # pylint: disable=too-many-arguments
    """ Obtain instance keypoints scores and coordinates
    """
    if (instance_keypoint_scores[source_keypoint_id] > 0.0 and
            instance_keypoint_scores[target_keypoint_id] == 0.0):
        source_keypoint = instance_keypoint_coords[source_keypoint_id]

        score, coords = _traverse_to_target_keypoint(edge,
                                                     source_keypoint,
                                                     target_keypoint_id,
                                                     scores,
                                                     offsets,
                                                     output_stride,
                                                     displacements)

        instance_keypoint_scores[target_keypoint_id] = score
        instance_keypoint_coords[target_keypoint_id] = coords


def _clip_to_indices(keypoints: np.ndarray,
                     output_stride: int,
                     width: int,
                     height: int) -> np.ndarray:
    """ Clip keypoint coordinate to indices within dimension (width, height)
    """
    keypoints = keypoints / output_stride
    keypoint_indices = np.zeros((2,), dtype=np.int32)

    keypoint_indices[0] = max(min(round(keypoints[0]), width - 1), 0)
    keypoint_indices[1] = max(min(round(keypoints[1]), height - 1), 0)

    return keypoint_indices


def _traverse_to_target_keypoint(edge_id: int,
                                 source_keypoint: np.ndarray,
                                 target_keypoint_id: int,
                                 scores: np.ndarray,
                                 offsets: np.ndarray,
                                 output_stride: int,
                                 displacements: np.ndarray) -> Tuple[float, float]:
    # pylint: disable=too-many-arguments
    """ Traverse to target keypoint to obtain keypoint score and coordinates
    """
    height = scores.shape[0] - 1
    width = scores.shape[1] - 1

    source_keypoint_indices = _clip_to_indices(
        source_keypoint, output_stride, width, height)

    displaced_point = source_keypoint + displacements[
        source_keypoint_indices[1], source_keypoint_indices[0], edge_id]

    displaced_point_indices = _clip_to_indices(
        displaced_point, output_stride, width, height)

    score = scores[displaced_point_indices[1],
                   displaced_point_indices[0], target_keypoint_id]

    image_coord = displaced_point_indices * output_stride + offsets[
        displaced_point_indices[1], displaced_point_indices[0], target_keypoint_id]

    return score, image_coord



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Displays info from dabble nodes such as fps, object count and zone counts in a legend box
"""

from typing import Any, Dict, List
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.draw.utils.legend import Legend


class Node(AbstractNode):
    """Draw node for drawing Legend box and info on image

    The draw legend node dynamically pulls the output results of previous nodes
    And uses it to draw the information into a legend box. Currently draws fps,
    object counts and object count in zones.

    all_legend_item: config is all possible items that can be drawn in legend box
    include: is used to select which information would be drawn
    This is so we can have the outputs but choose not to drawn on screen

    Inputs:

        all (:obj:`Any`): Receives inputs from all preceding outputs to use
        ass dynamic input for legend creation.

    Outputs:
        |none|

    Configs:
        position (:obj:`str`): **default = "bottom"**
            Position to draw legend box. "top" draws it at the top-left position
            while "bottom" draws it at bottom-left.

        include (:obj:`list`): **default = ["all_legend_items"]**
            List of information to draw. Current can draw "fps", "count" and/or
            "zone_count". The default value "all_legend_items" draws everything
            dynamically depending on inputs.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.legend_items: List[str] = []

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Draws legend box with information from nodes

        Args:
            inputs (dict): Dict with all available keys

        Returns:
            outputs (dict): Dict with keys "none"
        """
        if len(self.legend_items) == 0:
            # Check inputs to set legend items to draw
            if self.include[0] == 'all_legend_items': # type: ignore
                self.include = self.all_legend_items # type: ignore
            self._include(inputs)
        if len(self.legend_items) != 0:
            Legend().draw(inputs, self.legend_items, self.position)  # type: ignore
        else:
            return {}
        # cv2 weighted does not update the referenced image. Need to return and replace.
        return {'img': inputs['img']}

    def _include(self, inputs: Dict[str, Any]) -> None:
        for item in self.all_legend_items:
            if item in inputs.keys() and item in self.include:
                self.legend_items.append(item)



"""Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

from typing import Tuple, Any, Iterable, Union
import numpy as np
import cv2
from peekingduck.pipeline.nodes.draw.utils.constants import \
    CHAMPAGNE, THICK, POINT_RADIUS, TOMATO
from peekingduck.pipeline.nodes.draw.utils.general import \
    get_image_size, project_points_onto_original_image


def draw_human_poses(image: np.array,
                     keypoints: np.ndarray,
                     keypoint_conns: np.ndarray) -> None:
    # pylint: disable=too-many-arguments
    """Draw poses onto an image frame.

    Args:
        image (np.array): image of current frame
        keypoints (List[Any]): list of keypoint coordinates
        keypoints_conns (List[Any]): list of keypoint connections
    """
    image_size = get_image_size(image)
    num_persons = keypoints.shape[0]
    if num_persons > 0:
        for i in range(num_persons):
            _draw_connections(image, keypoint_conns[i],
                              image_size, CHAMPAGNE)
            _draw_keypoints(image, keypoints[i], image_size,
                            TOMATO, POINT_RADIUS)


def _draw_connections(frame: np.array,
                      connections: Union[None, Iterable[Any]],
                      image_size: Tuple[int, int],
                      connection_color: Tuple[int, int, int]) -> None:
    """ Draw connections between detected keypoints """
    if connections is not None:
        for connection in connections:
            pt1, pt2 = project_points_onto_original_image(connection, image_size)
            cv2.line(frame,
                     (pt1[0], pt1[1]),
                     (pt2[0], pt2[1]),
                     connection_color,
                     THICK)


def _draw_keypoints(frame: np.ndarray,
                    keypoints: np.ndarray,
                    image_size: Tuple[int, int],
                    keypoint_dot_color: Tuple[int, int, int],
                    keypoint_dot_radius: int) -> None:
    # pylint: disable=too-many-arguments
    """ Draw detected keypoints """
    img_keypoints = project_points_onto_original_image(
        keypoints, image_size)

    for _, keypoint in enumerate(img_keypoints):
        _draw_one_keypoint_dot(frame, keypoint, keypoint_dot_color, keypoint_dot_radius)


def _draw_one_keypoint_dot(frame: np.ndarray,
                           keypoint: np.ndarray,
                           keypoint_dot_color: Tuple[int, int, int],
                           keypoint_dot_radius: int) -> None:
    """ Draw single keypoint """
    cv2.circle(frame, (keypoint[0], keypoint[1]), keypoint_dot_radius, keypoint_dot_color, -1)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Pipeline class that stores nodes and manages the data information used during inference
"""

import textwrap
from typing import List, Dict, Any
from peekingduck.pipeline.nodes.node import AbstractNode


# pylint: disable=R0903
class Pipeline:
    """ Pipe class that stores nodes and manages flow of data used during inference

    Args:
        nodes (:obj: `List`): List of initiated nodes for the pipeline to run through

    Returns:
        None

    """

    def __init__(self, nodes: List[AbstractNode]) -> None:

        self.nodes = nodes
        self._check_pipe(nodes)
        self.data = {}  # type: ignore
        self.terminate = False

    def get_pipeline_results(self) -> Dict[str, Any]:
        """get all results data of nodes in pipeline

        Returns:
            Dict[Any]: Dictionary of all pipeline node results
        """
        return self.data

    @staticmethod
    def _check_pipe(nodes: List[AbstractNode]) -> None:
        # 1. Check the initial node is a source node
        # 2. Check every subsequent node utilizes something that will exist
        # if reached end, it is all valid

        data_pool = []

        if nodes[0].inputs[0] == 'none':
            data_pool.extend(nodes[0].outputs)

        for node in nodes[1:]:

            if all(item in data_pool for item in node.inputs) or "all" in node.inputs:
                data_pool.extend(node.outputs)
            else:
                msg = textwrap.dedent(f"""\
                    Nodes in this pipeline do not form a proper channel:
                    {node.name} requires these inputs: {node.inputs}
                    Data pool only has these outputs from previous nodes: {data_pool}
                    Note that nodes run sequentially, in the order specified in the config file.""")
                raise ValueError(msg)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helper classes to load configs, nodes
"""

import sys
import os
import importlib
import logging
import ast
import collections.abc
import re
from typing import Any, Dict, List

import yaml

from peekingduck.pipeline.pipeline import Pipeline
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.configloader import ConfigLoader

PEEKINGDUCK_NODE_TYPE = ["input", "model", "draw", "dabble", "output"]


class DeclarativeLoader:  # pylint: disable=too-few-public-methods
    """
    A helper class to create pipeline.

    The declarative loader class creates the specified nodes according to any
    modfications provided in the configs and returns the pipeline needed for
    inference.

    Args:

        run_config (:obj:`str`): Path to yaml file that declares the node \
        sequence to be used in the pipeline

        config_updates_cli (:obj:`str`): stringified nested dictionaries of \
        config changes passed as part of cli command. Used to modify the node \
        configurations direct from cli.

        custom_node_parent_folder (:obj:`str`): path to parent folder which contains \
        custom nodes that users have created to be used with PeekingDuck. \
        For more information on using custom nodes, please refer to \
        `getting started <getting_started/03_custom_nodes.html>`_.

    """

    def __init__(self,
                 run_config: str,
                 config_updates_cli: str,
                 custom_node_parent_folder: str) -> None:
        self.logger = logging.getLogger(__name__)

        pkdbasedir = os.path.join(os.path.dirname(os.path.abspath(__file__)))
        self.config_loader = ConfigLoader(pkdbasedir)

        self.node_list = self._load_node_list(run_config)
        self.config_updates_cli = ast.literal_eval(
            config_updates_cli)  # type: ignore

        custom_folder = self._get_custom_name_from_node_list()
        if custom_folder is not None:
            custom_folder_path = os.path.join(os.getcwd(),
                                              custom_node_parent_folder,
                                              custom_folder)
            self.custom_config_loader = ConfigLoader(custom_folder_path)
            sys.path.append(custom_node_parent_folder)

            self.custom_folder_path = custom_folder_path

    def _load_node_list(self, run_config: str) -> List[str]:
        """Loads a list of nodes from run_config.yml"""
        with open(run_config) as node_yml:
            nodes = yaml.safe_load(node_yml)['nodes']

        self.logger.info('Successfully loaded run_config file.')
        return nodes

    def _get_custom_name_from_node_list(self) -> Any:

        custom_name = None

        for node in self.node_list:
            if isinstance(node, dict) is True:
                node_type = [*node][0].split(".")[0]
            else:
                node_type = node.split(".")[0]

            if node_type not in PEEKINGDUCK_NODE_TYPE:
                custom_name = node_type
                break

        return custom_name

    def _instantiate_nodes(self) -> List[AbstractNode]:
        """ Given a list of imported nodes, instantiate nodes"""
        instantiated_nodes = []

        for node_item in self.node_list:
            config_updates_yml = None
            node_str = node_item

            if isinstance(node_item, dict):
                node_str = list(node_item.keys())[0]  # type: ignore
                config_updates_yml = node_item[node_str]

            node_str_split = node_str.split('.')

            msg = "Initialising " + node_str + " node..."
            self.logger.info(msg)

            if len(node_str_split) == 3:
                # convert windows/linux filepath to a module path
                path_to_node = ".".join(re.split(r'\\|\/', self.custom_folder_path)[-1:]) + "."
                node_name = ".".join(node_str_split[-2:])

                instantiated_node = self._init_node(path_to_node,
                                                    node_name,
                                                    self.custom_config_loader,
                                                    config_updates_yml)  # type: ignore
            else:
                path_to_node = 'peekingduck.pipeline.nodes.'

                instantiated_node = self._init_node(path_to_node,
                                                    node_str,
                                                    self.config_loader,
                                                    config_updates_yml)  # type: ignore

            instantiated_nodes.append(instantiated_node)

        return instantiated_nodes

    def _init_node(self, path_to_node: str, node_name: str,
                   config_loader: ConfigLoader,
                   config_updates_yml: Dict[str, Any]) -> AbstractNode:
        """ Import node to filepath and initialise node with config """

        node = importlib.import_module(path_to_node + node_name)
        config = config_loader.get(node_name)

        # First, override default configs with values from run_config.yml
        if config_updates_yml is not None:
            config = self._edit_config(config, config_updates_yml, node_name)

        # Second, override configs again with values from cli
        if self.config_updates_cli is not None:
            if node_name in self.config_updates_cli.keys():
                config = self._edit_config(
                    config, self.config_updates_cli[node_name], node_name)

        return node.Node(config)  # type: ignore

    def _edit_config(self, dict_orig: Dict[str, Any],
                     dict_update: Dict[str, Any],
                     node_name: str) -> Dict[str, Any]:
        """ Update value of a nested dictionary of varying depth using
            recursion
        """
        for key, value in dict_update.items():
            if isinstance(value, collections.abc.Mapping):
                dict_orig[key] = self._edit_config(
                    dict_orig.get(key, {}), value, node_name)  # type: ignore
            else:
                if key not in dict_orig:
                    self.logger.warning(
                        "Config for node %s does not have the key: %s",
                        node_name, key)
                else:
                    dict_orig[key] = value
                    self.logger.info(
                        "Config for node %s is updated to: '%s': %s",
                        node_name, key, value)
        return dict_orig

    def get_pipeline(self) -> Pipeline:
        """Returns a compiled Pipeline for PeekingDuck runner to execute"""
        instantiated_nodes = self._instantiate_nodes()

        try:
            return Pipeline(instantiated_nodes)
        except ValueError as error:
            self.logger.error(str(error))
            sys.exit(1)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Draws utilities regarding zoning functions
"""

from typing import List, Tuple, Any
import numpy as np
import cv2
from peekingduck.pipeline.nodes.draw.utils.constants import \
    PRIMARY_PALETTE, PRIMARY_PALETTE_LENGTH, VERY_THICK


def _draw_zone_area(frame: np.array, points: List[Tuple[int]],
                    zone_index: int) -> None:
    total_points = len(points)
    for i in range(total_points):
        if i == total_points-1:
            # for last point, link to first point
            cv2.line(frame, points[i], points[0],
                     PRIMARY_PALETTE[(zone_index+1) % PRIMARY_PALETTE_LENGTH], VERY_THICK)
        else:
            # for all other points, link to next point in polygon
            cv2.line(frame, points[i], points[i+1],
                     PRIMARY_PALETTE[(zone_index+1) % PRIMARY_PALETTE_LENGTH], VERY_THICK)


def draw_zones(frame: np.array, zones: List[Any]) -> None:
    """draw the boundaries of the zones used in zoning analytics

    Args:
        frame (np.array): image of current frame
        zones (Zone): zones used in the zoning analytics. possible
        classes are Area and Divider.
    """
    for i, zone_pts in enumerate(zones):
        _draw_zone_area(frame, zone_pts, i)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Code of this file is mostly forked from
# [@xuannianz](https://github.com/xuannianz))


"""
Anchor-related functions for EfficientDet
"""


from typing import Tuple, List, Callable
import numpy as np
from tensorflow import keras


class AnchorParameters:
    """
    The parameters that define how anchors are generated.

    Args
        sizes : List of sizes to use. Each size corresponds to one feature level.
        strides : List of strides to use. Each stride correspond to one feature level.
        ratios : List of ratios to use per location in a feature map.
        scales : List of scales to use per location in a feature map.
    """

    def __init__(self, sizes: Tuple[int, int, int, int, int] = (32, 64, 128, 256, 512),
                 strides: Tuple[int, int, int, int, int] = (8, 16, 32, 64, 128),
                 ratios: Tuple[float, float, float] = (1.0, 0.5, 2.0),
                 scales: Tuple[float, float, float] = (1.0, 2**(1. / 3.), 2**(2. / 3.))) -> None:
        self.sizes = sizes
        self.strides = strides
        self.ratios = np.array(ratios, dtype=keras.backend.floatx())
        self.scales = np.array(scales, dtype=keras.backend.floatx())

    def num_anchors(self) -> int:
        """Function to get number of anchors

        Returns:
            int: the total number of anchors
        """
        return len(self.ratios) * len(self.scales)

    def get_sizes(self) -> Tuple[int, int, int, int, int]:
        """Getter for sizes
        """
        return self.sizes


# The default anchor parameters.
AnchorParameters.default = AnchorParameters(  # type: ignore
    sizes=(32, 64, 128, 256, 512),
    strides=(8, 16, 32, 64, 128),
    ratios=np.array([1, 0.5, 2], keras.backend.floatx()),
    scales=np.array([2 ** 0, 2 ** (1.0 / 3.0), 2 ** (2.0 / 3.0)], keras.backend.floatx()),
)


def guess_shapes(image_shape: Tuple[int, int], pyramid_levels: List[int]) -> List[List]:
    """
    Guess shapes based on pyramid levels.

    Args
         image_shape: The shape of the image.
         pyramid_levels: A list of what pyramid levels are used.

    Returns
        A list of image shapes at each pyramid level.
    """
    image_shape = np.array(image_shape[:2])
    image_shapes = [(image_shape + 2 ** x - 1) // (2 ** x) for x in pyramid_levels]
    return image_shapes


def generate_anchors(base_size: int = 16,
                     ratios: Tuple[float] = None,
                     scales: Tuple[float] = None) -> np.ndarray:
    """
    Generate anchor windows by enumerating aspect ratios X scales w.r.t. a reference window.

    Args:
        base_size:
        ratios:
        scales:

    Returns:

    """
    if ratios is None:
        ratios = AnchorParameters.default.ratios  # type: ignore

    if scales is None:
        scales = AnchorParameters.default.scales  # type: ignore

    num_anchors = len(ratios) * len(scales)

    # initialize output anchors
    anchors = np.zeros((num_anchors, 4))

    anchors[:, 2:] = base_size * np.tile(np.repeat(scales, len(ratios))[None], (2, 1)).T

    areas = anchors[:, 2] * anchors[:, 3]

    # correct for ratios
    anchors[:, 2] = np.sqrt(areas / np.tile(ratios, len(scales)))
    anchors[:, 3] = anchors[:, 2] * np.tile(ratios, len(scales))

    anchors[:, 0::2] -= np.tile(anchors[:, 2] * 0.5, (2, 1)).T
    anchors[:, 1::2] -= np.tile(anchors[:, 3] * 0.5, (2, 1)).T

    return anchors


def shift(feature_map_shape: List[float], stride: int, anchors: np.ndarray) -> np.ndarray:
    """
    Produce shifted anchors based on shape of the map and stride size.

    Args
        feature_map_shape : Shape to shift the anchors over.
        stride : Stride to shift the anchors with over the shape.
        anchors: The anchors to apply at each location.
    """

    # create a grid starting from half stride from the top left corner
    shift_x = (np.arange(0, feature_map_shape[1]) + 0.5) * stride
    shift_y = (np.arange(0, feature_map_shape[0]) + 0.5) * stride

    shift_x, shift_y = np.meshgrid(shift_x, shift_y)

    shifts = np.vstack((
        shift_x.ravel(), shift_y.ravel(),
        shift_x.ravel(), shift_y.ravel()
    )).transpose()

    a_num = anchors.shape[0]
    k_num = shifts.shape[0]
    all_anchors = (anchors.reshape((1, a_num, 4)) +
                   shifts.reshape((1, k_num, 4)).transpose((1, 0, 2)))
    all_anchors = all_anchors.reshape((k_num * a_num, 4))

    return all_anchors


def anchors_for_shape(
        image_shape: Tuple[int, int],
        pyramid_levels: List[int] = None,
        anchor_params: AnchorParameters = None,
        shapes_callback: Callable = None,
) -> np.ndarray:
    """
    Generators anchors for a given shape.

    Args
        image_shape: The shape of the image.
        pyramid_levels: List representing which pyramids to use (defaults to [3, 4, 5, 6, 7]).
        anchor_params: Struct containing anchor parameters. If None, default values are used.
        shapes_callback: Function to call to get the shape of image at different pyramid levels.

    Returns
        np.array of shape (N, 4) containing the (x1, y1, x2, y2) coordinates for the anchors.
    """

    if pyramid_levels is None:
        pyramid_levels = [3, 4, 5, 6, 7]

    if anchor_params is None:
        anchor_params = AnchorParameters.default  # type: ignore

    if shapes_callback is None:
        shapes_callback = guess_shapes
    feature_map_shapes = shapes_callback(image_shape, pyramid_levels)

    # compute anchors over all pyramid levels
    all_anchors = np.zeros((0, 4), dtype=np.float32)
    for idx, _ in enumerate(pyramid_levels):
        anchors = generate_anchors(
            base_size=anchor_params.sizes[idx],
            ratios=anchor_params.ratios,
            scales=anchor_params.scales
        )
        shifted_anchors = shift(feature_map_shapes[idx], anchor_params.strides[idx], anchors)
        all_anchors = np.append(all_anchors, shifted_anchors, axis=0)

    return all_anchors.astype(np.float32)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Loads configurations for individual nodes
"""

import os
from typing import Any, Dict

import yaml


class ConfigLoader:  # pylint: disable=too-few-public-methods
    """
    A helper class to create pipeline.

    The config loader class is used to allow for instantiation of Node classes
    directly instead of reading configurations from the run config yaml.

    Args:

        basedir (:obj:`str`): base directory of peekingduck

    """

    def __init__(self, basedir: str) -> None:
        self._basedir = basedir

    def _get_config_path(self, node: str) -> str:
        """ Based on the node, return the corresponding node config path """

        configs_folder = os.path.join(self._basedir, 'configs')
        node_type, node_name = node.split(".")
        node_name = node_name + ".yml"
        filepath = os.path.join(configs_folder, node_type, node_name)

        return filepath

    def get(self, node_name: str) -> Dict[str, Any]:
        """
        Get node configuration for specified node.

        Args:

            node_name (:obj:`str`): name of node

        Outputs:
            node_config (:obj:`Dict`): dictionary of node configurations for the
            specified node

        """

        filepath = self._get_config_path(node_name)

        with open(filepath) as file:
            node_config = yaml.safe_load(file)

        # some models require the knowledge of where the root is for loading
        node_config['root'] = self._basedir
        return node_config



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Code of this file is mostly forked from
# [@xuannianz](https://github.com/xuannianz))

"""
Inject Tensorflow Keras into model workloads
"""


from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files.utils.keras_utils \
    import inject_tfkeras_modules, init_tfkeras_custom_objects
from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files \
    import efficientnet as model

EfficientNetB0 = inject_tfkeras_modules(model.efficientnet_b0)
EfficientNetB1 = inject_tfkeras_modules(model.efficientnet_b1)
EfficientNetB2 = inject_tfkeras_modules(model.efficientnet_b2)
EfficientNetB3 = inject_tfkeras_modules(model.efficientnet_b3)
EfficientNetB4 = inject_tfkeras_modules(model.efficientnet_b4)
EfficientNetB5 = inject_tfkeras_modules(model.efficientnet_b5)
EfficientNetB6 = inject_tfkeras_modules(model.efficientnet_b6)
EfficientNetB7 = inject_tfkeras_modules(model.efficientnet_b7)

preprocess_input = inject_tfkeras_modules(model.preprocess_input)

init_tfkeras_custom_objects()



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Abstract Node class for all nodes
"""

import pathlib
import logging
import collections
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List

from peekingduck.configloader import ConfigLoader


class AbstractNode(metaclass=ABCMeta):
    """
    Abstract Node class for inheritance by nodes.
    It defines default attributes and methods of a node.
    """

    def __init__(self, config: Dict[str, Any] = None,
                 node_path: str = "", pkdbasedir: str = None, **kwargs: Any) -> None:

        self._name = node_path
        self.logger = logging.getLogger(self._name)

        if not pkdbasedir:
            pkdbasedir = str(pathlib.Path(__file__).parents[2].resolve())

        self.node_name = ".".join(node_path.split(".")[-2:])

        # NOTE: config and kwargs_config are similar but are from different inputs
        # config is when users input a dictionary to update the node
        # kwargs_config is when users input parameters to update the node
        self.config_loader = ConfigLoader(pkdbasedir)
        self.load_node_config(config, kwargs)  # type: ignore

    @classmethod
    def __subclasshook__(cls: Any, subclass: Any) -> bool:
        return (hasattr(subclass, 'run') and
                callable(subclass.run))

    @abstractmethod
    def run(self, inputs: Dict[str, Any]) -> None:
        """abstract method needed for running node"""
        raise NotImplementedError("This method needs to be implemented")

    @property
    def inputs(self) -> List[str]:
        """getter for input requirements"""
        return self.input

    @property
    def outputs(self) -> List[str]:
        """getter for node outputs"""
        return self.output

    @property
    def name(self) -> str:
        """getter for node name"""
        return self._name

    def load_node_config(self,
                         config: Dict[str, Any],
                         kwargs_config: Dict[str, Any]) -> None:
        """ loads node configuration
        NOTE: config and kwargs_config are similar but come from different inputs
        config is when users input a dictionary to update the node
        kwargs_config is when users input parameters to update the node

        Args:
            config (Dict[str, Any]): loads configuration from a dictionary input
            kwargs_config (Dict[str, Any]): loads configuration from kwargs
        """

        # if full set of configuration is not included in config
        # load configuration and update node with **kwargs where possible
        # else load from kwargs only
        self.config = config
        if not self.config:
            loaded_config = self.config_loader.get(self.node_name)
            updated_config = self._edit_config(loaded_config, kwargs_config)
            self.config = updated_config

        # sets class attributes
        for key in self.config:
            setattr(self, key, self.config[key])

    # pylint: disable=R0801
    def _edit_config(self,
                     dict_orig: Dict[str, Any],
                     dict_update: Dict[str, Any]) -> Dict[str, Any]:
        """ Update value of a nested dictionary of varying depth using
            recursion
        """
        if dict_update:
            for key, value in dict_update.items():
                if isinstance(value, collections.abc.Mapping):
                    dict_orig[key] = self._edit_config(
                        dict_orig.get(key, {}), value)  # type: ignore
                elif key not in dict_orig:
                    self.logger.warning(
                        "Config for node %s does not have the key: %s",
                        self.node_name, key)
                else:
                    dict_orig[key] = value
                    self.logger.info(
                        "Config for node %s is updated to: '%s': %s",
                        self.node_name, key, value)
        return dict_orig



# Copyright 2021 AI Singapore

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#      https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Code of this file is mostly forked from
# [@xuannianz](https://github.com/xuannianz))

"""
Utility functions to obtain submodules
"""

from typing import Dict, Any, Tuple

_KERAS_BACKEND = None
_KERAS_LAYERS = None
_KERAS_MODELS = None
_KERAS_UTILS = None


def get_submodules_from_kwargs(kwargs: Dict[str, Any]) -> Tuple[Any, Any, Any, Any]:
    """Helper function to get keras submodules

    Args:
        kwargs (dict]): dictionary of keyword args

    Returns:
        (tuple): the specified submodules
    """
    backend = kwargs.get('backend', _KERAS_BACKEND)
    layers = kwargs.get('layers', _KERAS_LAYERS)
    models = kwargs.get('models', _KERAS_MODELS)
    utils = kwargs.get('utils', _KERAS_UTILS)
    for key in kwargs.keys():
        if key not in ['backend', 'layers', 'models', 'utils']:
            raise TypeError(f'Invalid keyword argument: {key}')
    return backend, layers, models, utils



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Slower, accurate Object Detection model
"""


from typing import Dict, Any
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.model.efficientdet_d04 import efficientdet_model


class Node(AbstractNode):
    """EfficientDet node that initalises a EfficientDet model to detect bounding
    boxes from an image.

    The EfficientDet node is capable of detecting objects from 80 categories.
    The table of categories can be found :term:`here <object detection
    indices>`.

    EfficientDet node has five levels of compound coefficient
    (0 - 5). A higher compound coefficient will scale up all dimensions of the
    backbone network width, depth, and input resolution, which results in better
    performance but slower inference time. The default compound coefficient is
    0 and can be changed to other values.

    Inputs:
        |img|

    Outputs:
        |bboxes|

        |bbox_labels|

        |bbox_scores|

    Configs:
        model_type (:obj:`int`): **{"0", "1", "2", "3", "4"}, default="0"**

            Defines the compound coefficient for EfficientDet.

        score_threshold (:obj:`float`): **[0,1], default = 0.3**

            Threshold to determine if detection should be returned

        detect_ids (:obj:`List`): **default = [0] **

            List of object class ids to be detected.

    References:

    EfficientDet: Scalable and Efficient Object Detection: https://arxiv.org/abs/1911.09070

    Code adapted from https://github.com/xuannianz/EfficientDet
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.model = efficientdet_model.EfficientDetModel(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Function that takes an image as input and returns bboxes of objects specified
        in config.
        """
        # Currently prototyped to return just the bounding boxes
        # without the scores
        bboxes, labels, scores = self.model.predict(inputs["img"])
        outputs = {"bboxes": bboxes, "bbox_labels": labels, "bbox_scores": scores}
        return outputs



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Code of this file is mostly forked from
# [@xuannianz](https://github.com/xuannianz))

"""
Processing helper functinos for EfficientDet
"""

from typing import List, Tuple
import numpy as np
import cv2

IMG_MEAN = [0.485, 0.456, 0.406]
IMG_STD = [0.229, 0.224, 0.225]


def preprocess_image(image: np.ndarray,
                     image_size: int) -> Tuple[List[List[float]], float]:
    """Preprocessing helper function for efficientdet

    Args:
        image (np.array): the input image in numpy array
        image_size (int): the model input size as specified in config

    Returns:
        image (np.array): the preprocessed image
        scale (float): the scale in which the original image was resized to
    """
    # image, RGB
    image_height, image_width = image.shape[:2]
    if image_height > image_width:
        scale = image_size / image_height
        resized_height = image_size
        resized_width = int(image_width * scale)
    else:
        scale = image_size / image_width
        resized_height = int(image_height * scale)
        resized_width = image_size

    image = cv2.resize(image, (resized_width, resized_height))
    image = image.astype(np.float32)
    image /= 255.
    image -= IMG_MEAN
    image /= IMG_STD
    pad_h = image_size - resized_height
    pad_w = image_size - resized_width
    image = np.pad(image, [(0, pad_h), (0, pad_w), (0, 0)], mode='constant')

    return image, scale


def postprocess_boxes(boxes: np.ndarray,
                      scale: float,
                      height: int,
                      width: int) -> np.ndarray:
    """Postprocessing helper function for efficientdet

    Args:
        boxes (np.array): the original detected bboxes from model output
        scale (float): scale in which the original image was resized to
        height (int): the height of the original image
        width (int): the width of the original image

    Returns:
        boxes (np.array): the postprocessed bboxes
    """
    boxes /= scale
    boxes[:, 0] = np.clip(boxes[:, 0], 0, width - 1)
    boxes[:, 1] = np.clip(boxes[:, 1], 0, height - 1)
    boxes[:, 2] = np.clip(boxes[:, 2], 0, width - 1)
    boxes[:, 3] = np.clip(boxes[:, 3], 0, height - 1)

    boxes[:, [0, 2]] /= width
    boxes[:, [1, 3]] /= height
    return boxes



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
functions for drawing legend related UI components
"""

from typing import Dict, List, Any
import numpy as np
import cv2
from cv2 import FONT_HERSHEY_SIMPLEX, LINE_AA
from peekingduck.pipeline.nodes.draw.utils.constants import \
    THICK, WHITE, SMALL_FONTSCALE, BLACK, FILLED, \
    PRIMARY_PALETTE, PRIMARY_PALETTE_LENGTH
from peekingduck.pipeline.nodes.draw.utils.general import get_image_size


class Legend:
    """Legend class that uses available info to draw legend box on frame"""

    def __init__(self) -> None:
        self.func_reg = self._get_legend_registry()
        self.legend_left_x = 15

        self.frame = None
        self.legend_starting_y = 0
        self.delta_y = 0
        self.legend_height = 0

    def draw(self, inputs: Dict[str, Any], items: List[str], position: str) -> np.array:
        """ Draw legends onto image

        Args:
            inputs (dict): dictionary of all available inputs for drawing the legend
            items (list): list of items to be drawn in legend
            position (str): used to control whether legend box is drawn on top or bottom
        """
        self.frame = inputs['img']

        self.legend_height = self._get_legend_height(inputs, items)
        self._set_legend_variables(position)

        self._draw_legend_box(self.frame)
        y_pos = self.legend_starting_y + 20
        for item in items:
            self.func_reg[item](self.frame, y_pos, inputs[item])
            y_pos += 20

    def _draw_count(self, frame: np.array, y_pos: int, count: int) -> None:
        """draw count of selected object onto frame

        Args:
            frame (np.array): image of current frame
            y_pos (int): y_position to draw the count text
            count (int): total count of selected object
                in current frame
        """
        text = 'COUNT: {0}'.format(count)
        cv2.putText(frame, text, (self.legend_left_x + 10, y_pos), FONT_HERSHEY_SIMPLEX,
                    SMALL_FONTSCALE, WHITE, THICK, LINE_AA)

    def _draw_fps(self, frame: np.array, y_pos: int, current_fps: float) -> None:
        """ Draw FPS onto frame image

        Args:
            frame (np.array): image of current frame
            y_pos (int): y position to draw the count info text
            current_fps (float): value of the calculated FPS
        """
        text = "FPS: {:.05}".format(current_fps)

        cv2.putText(frame, text, (self.legend_left_x + 10, y_pos), FONT_HERSHEY_SIMPLEX,
                    SMALL_FONTSCALE, WHITE, THICK, LINE_AA)

    def _draw_zone_count(self, frame:np.array, y_pos: int, counts: List[int]) -> None:
        """ Draw zone counts of all zones onto frame image

        Args:
            frame (np.array): image of current frame
            y_pos (int): y position to draw the count info text
            counts (list): list of zone counts
        """
        text = '-ZONE COUNTS-'
        cv2.putText(frame, text, (self.legend_left_x + 10, y_pos), FONT_HERSHEY_SIMPLEX,
                    SMALL_FONTSCALE, WHITE, THICK, LINE_AA)
        for i, count in enumerate(counts):
            y_pos += 20
            cv2.rectangle(frame,
                        (self.legend_left_x + 10, y_pos - 14),
                        (self.legend_left_x + 30, y_pos + 5),
                        PRIMARY_PALETTE[(i+1) % PRIMARY_PALETTE_LENGTH],
                        FILLED)
            text = ' ZONE-{0}: {1}'.format(i+1, count)
            cv2.putText(frame,
                        text,
                        (40, y_pos),
                        FONT_HERSHEY_SIMPLEX,
                        SMALL_FONTSCALE,
                        WHITE,
                        THICK,
                        LINE_AA)

    def _draw_legend_box(self, frame: np.array) -> None:
        """draw pts of selected object onto frame

        Args:
            frame (np.array): image of current frame
            zone_count (List[float]): object count, likely people, of each zone used
            in the zone analytics
        """
        assert self.legend_height is not None
        overlay = frame.copy()
        cv2.rectangle(overlay,
                      (self.legend_left_x, self.legend_starting_y),
                      (self.legend_left_x + 150,
                       self.legend_starting_y + self.legend_height),
                      BLACK,
                      FILLED)
        # apply the overlay
        cv2.addWeighted(overlay, 0.3, frame, 0.7, 0, frame)

    @staticmethod
    def _get_legend_height(inputs: Dict[str, Any], items: List[str]) -> int:
        """Get height of legend box needed to contain all items drawn"""
        no_of_items = len(items)
        if "zone_count" in items:
            # increase the number of items according to number of zones
            no_of_items += len(inputs["zone_count"])
        return 12 * no_of_items + 8 * (no_of_items - 1) + 20

    def _get_legend_registry(self) -> Dict[str, Any]:
        """Get registry of functions that draw items
        available in the legend"""
        return {
            'fps': self._draw_fps,
            'count': self._draw_count,
            'zone_count': self._draw_zone_count
        }

    def _set_legend_variables(self, position: str) -> None:
        assert self.legend_height != 0
        if position == "top":
            self.legend_starting_y = 10
        else:
            _, image_height = get_image_size(self.frame)
            self.legend_starting_y = image_height - 10 - self.legend_height

    def add_register(self, name: str, method: Any) -> None:
        """Add new legend drawing information to the registry

        Args:
            name (str): name of method, corresponding to key to get input
            method (Any): function of the method. Note that take in 1) image frame
            and 2) input needed for the method, taken in using inputs[<key>]
        """
        self.func_reg[name] = method



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Detector class to handle detection of poses for hrnet
"""

import os
import logging
from typing import Any, Dict, List, Tuple, Callable

import tensorflow as tf
import numpy as np
from peekingduck.utils.graph_functions import load_graph
from peekingduck.pipeline.nodes.model.hrnetv1.hrnet_files.preprocessing \
    import project_bbox, box2cs, crop_and_resize
from peekingduck.pipeline.nodes.model.hrnetv1.hrnet_files.postprocessing \
    import scale_transform, affine_transform_xy, reshape_heatmaps, \
    get_valid_keypoints, get_keypoint_conns


class Detector:
    """Detector class to handle detection of poses for hrnet
    """

    def __init__(self, config: Dict[str, Any]) -> None:
        self.logger = logging.getLogger(__name__)

        self.config = config
        self.root_dir = config['root']
        self.resolution = config['resolution']
        self.min_score = config['score_threshold']

        self.hrnet = self._create_hrnet_model()

    def _inference_function(self, person_frame: np.ndarray, training: bool = False) -> tf.Tensor:  # pylint:disable=unused-argument
        '''
        When graph is frozen, we need a different way to extract the
        arrays. We use this to return the values needed. The purpose
        of this function is to make it consistent with how our regular
        hrnet is used for inference.

        The "training" argument is needed because in the usual implementation
        of hrnet has the training argument. It does nothing here.
        '''
        heatmap = self.frozen_fn(tf.cast(person_frame, float))[0]
        return heatmap

    def _create_hrnet_model(self) -> Callable:
        graph_path = os.path.join(self.root_dir, self.config['model_file'])
        model_nodes = self.config['MODEL_NODES']
        self.frozen_fn = load_graph(graph_path,
                                    inputs=model_nodes['inputs'],
                                    outputs=model_nodes['outputs'])
        resolution_tuple = (self.resolution['height'], self.resolution['width'])
        self.logger.info(
            'HRNet graph model loaded with following configs: \n \
            Resolution: %s, \n \
            Score Threshold: %s', resolution_tuple, self.min_score)
        return self._inference_function

    def preprocess(self,
                   frame: np.ndarray,
                   bboxes: np.ndarray) -> Tuple[np.ndarray, np.ndarray, Tuple[int, int]]:
        """Preprocessing function that crops bboxes while preserving aspect ratio

        Args:
            frame (np.ndarray): input image in numpy array
            bboxes (np.ndarray): array of detected bboxes

        Returns:
            Tuple[np.ndarray, np.ndarray, Tuple[int, int]]: array of cropped bboxes, \
            transformation matrices and original frame size
        """
        frame = frame / 255.0
        frame_size = (frame.shape[1], frame.shape[0])
        cropped_size = (self.resolution['width'], self.resolution['height'])

        projected_bbox = project_bbox(bboxes, frame_size)
        center_bbox = box2cs(projected_bbox, self.resolution['width'] / self.resolution['height'])
        cropped_imgs, affine_matrices = crop_and_resize(frame, center_bbox, cropped_size)

        return np.array(cropped_imgs), affine_matrices, frame_size

    def postprocess(self, heatmaps: np.ndarray,  # pylint: disable=too-many-locals
                    affine_matrices: np.ndarray,
                    cropped_frames_scale: List[int],
                    frame_size: Tuple[int, int]) -> Tuple[np.ndarray,
                                                          np.ndarray,
                                                          np.ndarray]:
        """Postprocessing function to process output heatmaps to required keypoint arays

        Args:
            heatmaps (np.ndarray): Output heatmaps from hrnet network
            affine_matrices (np.ndarray): transformation matrices of preprocess cropping
            cropped_frames_scale (List[int]): Shape of cropped bboxes
            frame_size (Tuple[int, int]): Size of original image

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray]: \
            tuple containing array of bboxes and pose related info i.e coordinates,
            scores, connections
        """
        batch, out_h, out_w, num_joints = heatmaps.shape
        heatmaps_reshaped = reshape_heatmaps(heatmaps)

        max_idxs = np.argmax(heatmaps_reshaped, 2)
        kp_scores = np.amax(heatmaps_reshaped, 2)
        keypoints = np.repeat(max_idxs, 2).reshape(batch, num_joints, -1).astype(np.float32)
        keypoints[:, :, 0] = (keypoints[:, :, 0]) % out_w
        keypoints[:, :, 1] = np.floor((keypoints[:, :, 1]) / out_w)

        keypoints = scale_transform(keypoints,
                                    in_scale=[out_w, out_h],
                                    out_scale=cropped_frames_scale)
        keypoints = affine_transform_xy(keypoints, affine_matrices)
        keypoints, kp_masks = get_valid_keypoints(
            keypoints, kp_scores, batch, self.min_score)

        normalized_keypoints = keypoints / frame_size

        keypoint_conns = get_keypoint_conns(normalized_keypoints, kp_masks)

        return normalized_keypoints, kp_scores, keypoint_conns

    def predict(self, frame: np.ndarray, bboxes: np.ndarray) -> Tuple[np.ndarray,
                                                                      np.ndarray,
                                                                      np.ndarray]:
        """HRnet prediction function

        Args:
            frame (np.ndarray): Image in numpy array
            bboxes (np.ndarray): Array of detected bboxes

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray]: \
            tuple containing list of bboxes and pose related info i.e coordinates,
            scores, connections
        """
        cropped_frames, affine_matrices, frame_size = self.preprocess(frame, bboxes)
        heatmaps = self.hrnet(cropped_frames, training=False)

        cropped_frames_scale = [cropped_frames.shape[2], cropped_frames.shape[1]]
        poses, kp_scores, kp_conns = self.postprocess(heatmaps.numpy(), affine_matrices,
                                                      cropped_frames_scale, frame_size)

        return poses, kp_scores, kp_conns



# Copyright 2019 The TensorFlow Authors, Pavel Yakubovskiy, Björn Barz. All Rights Reserved.
#
# Modifications copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# Contains definitions for EfficientNet model.
#
# [1] Mingxing Tan, Quoc V. Le
#   EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks.
#   ICML'19, https://arxiv.org/abs/1905.11946
#
# Code of this model implementation is mostly written by
# Björn Barz ([@Callidior](https://github.com/Callidior))

"""
EfficientNet parameters and constants
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import math
import string
import collections
from typing import Any, Dict, Callable, List, Tuple, Union

from six.moves import xrange
import tensorflow as tf
from tensorflow.python.keras.applications.imagenet_utils import obtain_input_shape
from tensorflow.keras.applications.imagenet_utils import preprocess_input as _preprocess_input

from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files.utils.submodule \
    import get_submodules_from_kwargs


BACKEND = None
LAYERS = None
MODELS = None
KERAS_UTILS = None


BASE_WEIGHTS_PATH = (
    'https://github.com/Callidior/keras-applications/'
    'releases/download/efficientnet/')

WEIGHTS_HASHES = {
    'efficientnet-b0': ('163292582f1c6eaca8e7dc7b51b01c61'
                        '5b0dbc0039699b4dcd0b975cc21533dc',
                        'c1421ad80a9fc67c2cc4000f666aa507'
                        '89ce39eedb4e06d531b0c593890ccff3'),
    'efficientnet-b1': ('d0a71ddf51ef7a0ca425bab32b7fa7f1'
                        '6043ee598ecee73fc674d9560c8f09b0',
                        '75de265d03ac52fa74f2f510455ba64f'
                        '9c7c5fd96dc923cd4bfefa3d680c4b68'),
    'efficientnet-b2': ('bb5451507a6418a574534aa76a91b106'
                        'f6b605f3b5dde0b21055694319853086',
                        '433b60584fafba1ea3de07443b74cfd3'
                        '2ce004a012020b07ef69e22ba8669333'),
    'efficientnet-b3': ('03f1fba367f070bd2545f081cfa7f3e7'
                        '6f5e1aa3b6f4db700f00552901e75ab9',
                        'c5d42eb6cfae8567b418ad3845cfd63a'
                        'a48b87f1bd5df8658a49375a9f3135c7'),
    'efficientnet-b4': ('98852de93f74d9833c8640474b2c698d'
                        'b45ec60690c75b3bacb1845e907bf94f',
                        '7942c1407ff1feb34113995864970cd4'
                        'd9d91ea64877e8d9c38b6c1e0767c411'),
    'efficientnet-b5': ('30172f1d45f9b8a41352d4219bf930ee'
                        '3339025fd26ab314a817ba8918fefc7d',
                        '9d197bc2bfe29165c10a2af8c2ebc675'
                        '07f5d70456f09e584c71b822941b1952'),
    'efficientnet-b6': ('f5270466747753485a082092ac9939ca'
                        'a546eb3f09edca6d6fff842cad938720',
                        '1d0923bb038f2f8060faaf0a0449db4b'
                        '96549a881747b7c7678724ac79f427ed'),
    'efficientnet-b7': ('876a41319980638fa597acbbf956a82d'
                        '10819531ff2dcb1a52277f10c7aefa1a',
                        '60b56ff3a8daccc8d96edfd40b204c11'
                        '3e51748da657afd58034d54d3cec2bac')
}

BlockArgs = collections.namedtuple('BlockArgs', [
    'kernel_size', 'num_repeat', 'input_filters', 'output_filters',
    'expand_ratio', 'id_skip', 'strides', 'se_ratio'
])
# defaults will be a public argument for namedtuple in Python 3.7
# https://docs.python.org/3/library/collections.html#collections.namedtuple
# mypy: ignore-errors
BlockArgs.__new__.__defaults__ = (None,) * len(BlockArgs._fields)  # type: ignore

DEFAULT_BLOCKS_ARGS = (
    BlockArgs(kernel_size=3, num_repeat=1, input_filters=32, output_filters=16,
              expand_ratio=1, id_skip=True, strides=[1, 1], se_ratio=0.25),
    BlockArgs(kernel_size=3, num_repeat=2, input_filters=16, output_filters=24,
              expand_ratio=6, id_skip=True, strides=[2, 2], se_ratio=0.25),
    BlockArgs(kernel_size=5, num_repeat=2, input_filters=24, output_filters=40,
              expand_ratio=6, id_skip=True, strides=[2, 2], se_ratio=0.25),
    BlockArgs(kernel_size=3, num_repeat=3, input_filters=40, output_filters=80,
              expand_ratio=6, id_skip=True, strides=[2, 2], se_ratio=0.25),
    BlockArgs(kernel_size=5, num_repeat=3, input_filters=80, output_filters=112,
              expand_ratio=6, id_skip=True, strides=[1, 1], se_ratio=0.25),
    BlockArgs(kernel_size=5, num_repeat=4, input_filters=112, output_filters=192,
              expand_ratio=6, id_skip=True, strides=[2, 2], se_ratio=0.25),
    BlockArgs(kernel_size=3, num_repeat=1, input_filters=192, output_filters=320,
              expand_ratio=6, id_skip=True, strides=[1, 1], se_ratio=0.25)
)

CONV_KERNEL_INITIALIZER = {
    'class_name': 'VarianceScaling',
    'config': {
        'scale': 2.0,
        'mode': 'fan_out',
        # EfficientNet actually uses an untruncated normal distribution for
        # initializing conv layers, but keras.initializers.VarianceScaling use
        # a truncated distribution.
        # We decided against a custom initializer for better serializability.
        'distribution': 'normal'
    }
}

DENSE_KERNEL_INITIALIZER = {
    'class_name': 'VarianceScaling',
    'config': {
        'scale': 1. / 3.,
        'mode': 'fan_out',
        'distribution': 'uniform'
    }
}


def preprocess_input(model_input: tf.Tensor, **kwargs: Dict[str, Any]) -> tf.Tensor:
    """Preprocesses model input
    """
    kwargs = {k: v for k, v in kwargs.items() if k in ['backend', 'layers', 'models', 'utils']}
    return _preprocess_input(model_input, mode='torch', **kwargs)


def get_swish(**kwargs: Dict[str, Any]) -> Callable:
    """Swish activation function: x * sigmoid(x).
        Reference: [Searching for Activation Functions](https://arxiv.org/abs/1710.05941)
        """
    backend, _, _, _ = get_submodules_from_kwargs(kwargs)

    def swish(swish_x: tf.Tensor) -> tf.Tensor:

        if backend.backend() == 'tensorflow':
            try:
                # The native TF implementation has a more
                # memory-efficient gradient implementation
                return backend.tf.nn.swish(swish_x)
            except AttributeError:
                pass

        return swish_x * backend.sigmoid(swish_x)

    return swish


def get_dropout(**kwargs: Union[None, Dict[str, Any]]) -> Any:
    """Wrapper over custom dropout. Fix problem of ``None`` shape for tf.keras.
    It is not possible to define FixedDropout class as global object,
    because we do not have modules for inheritance at first time.

    Issue:
        https://github.com/tensorflow/tensorflow/issues/30946
    """
    backend, layers, _, _ = get_submodules_from_kwargs(kwargs)

    class FixedDropout(layers.Dropout):  # pylint: disable=too-few-public-methods
        """Fixed Dropout Class
        """

        def _get_noise_shape(self, inputs: tf.Tensor) -> Union[None, Tuple[Any, ...]]:
            if self.noise_shape is None:
                return self.noise_shape

            symbolic_shape = backend.shape(inputs)
            noise_shape = [symbolic_shape[axis] if shape is None else shape
                           for axis, shape in enumerate(self.noise_shape)]
            return tuple(noise_shape)

    return FixedDropout


def round_filters(filters: float, width_coefficient: float, depth_divisor: int) -> int:
    """Round number of filters based on width multiplier."""

    filters *= width_coefficient
    new_filters = int(filters + depth_divisor / 2) // depth_divisor * depth_divisor
    new_filters = max(depth_divisor, new_filters)
    # Make sure that round down does not go down by more than 10%.
    if new_filters < 0.9 * filters:
        new_filters += depth_divisor
    return int(new_filters)


def round_repeats(repeats: int, depth_coefficient: float) -> int:
    """Round number of repeats based on depth multiplier."""

    return int(math.ceil(depth_coefficient * repeats))


def mb_conv_block(inputs: tf.Tensor,  # type:ignore
                  block_args,
                  activation: Callable,
                  drop_rate: float = None,
                  prefix: str = '') -> tf.Tensor:
    """Mobile Inverted Residual Bottleneck."""

    has_se = (block_args.se_ratio is not None) and (0 < block_args.se_ratio <= 1)
    bn_axis = 3 if BACKEND.image_data_format() == 'channels_last' else 1

    # workaround over non working dropout with None in noise_shape in tf.keras
    Dropout = get_dropout(
        backend=BACKEND,
        layers=LAYERS,
        models=MODELS,
        utils=KERAS_UTILS
    )

    # Expansion phase
    filters = block_args.input_filters * block_args.expand_ratio
    if block_args.expand_ratio != 1:
        x_in = LAYERS.Conv2D(filters, 1,
                             padding='same',
                             use_bias=False,
                             kernel_initializer=CONV_KERNEL_INITIALIZER,
                             name=prefix + 'expand_conv')(inputs)
        x_in = LAYERS.BatchNormalization(axis=bn_axis, name=prefix + 'expand_bn')(x_in)
        x_in = LAYERS.Activation(activation, name=prefix + 'expand_activation')(x_in)
    else:
        x_in = inputs

    # Depthwise Convolution
    x_in = LAYERS.DepthwiseConv2D(block_args.kernel_size,
                                  strides=block_args.strides,
                                  padding='same',
                                  use_bias=False,
                                  depthwise_initializer=CONV_KERNEL_INITIALIZER,
                                  name=prefix + 'dwconv')(x_in)
    x_in = LAYERS.BatchNormalization(axis=bn_axis, name=prefix + 'bn')(x_in)
    x_in = LAYERS.Activation(activation, name=prefix + 'activation')(x_in)

    # Squeeze and Excitation phase
    if has_se:
        num_reduced_filters = max(1, int(
            block_args.input_filters * block_args.se_ratio
        ))
        se_tensor = LAYERS.GlobalAveragePooling2D(name=prefix + 'se_squeeze')(x_in)

        target_shape = (1, 1, filters) if BACKEND.image_data_format(
        ) == 'channels_last' else (filters, 1, 1)
        se_tensor = LAYERS.Reshape(target_shape, name=prefix + 'se_reshape')(se_tensor)
        se_tensor = LAYERS.Conv2D(num_reduced_filters, 1,
                                  activation=activation,
                                  padding='same',
                                  use_bias=True,
                                  kernel_initializer=CONV_KERNEL_INITIALIZER,
                                  name=prefix + 'se_reduce')(se_tensor)
        se_tensor = LAYERS.Conv2D(filters, 1,
                                  activation='sigmoid',
                                  padding='same',
                                  use_bias=True,
                                  kernel_initializer=CONV_KERNEL_INITIALIZER,
                                  name=prefix + 'se_expand')(se_tensor)
        if BACKEND.backend() == 'theano':
            # For the Theano backend, we have to explicitly make
            # the excitation weights broadcastable.
            pattern = ([True, True, True, False] if BACKEND.image_data_format() == 'channels_last'
                       else [True, False, True, True])
            se_tensor = LAYERS.Lambda(
                lambda x: BACKEND.pattern_broadcast(x, pattern),
                name=prefix + 'se_broadcast')(se_tensor)
        x_in = LAYERS.multiply([x_in, se_tensor], name=prefix + 'se_excite')

    # Output phase
    x_in = LAYERS.Conv2D(block_args.output_filters, 1,
                         padding='same',
                         use_bias=False,
                         kernel_initializer=CONV_KERNEL_INITIALIZER,
                         name=prefix + 'project_conv')(x_in)
    x_in = LAYERS.BatchNormalization(axis=bn_axis, name=prefix + 'project_bn')(x_in)
    if block_args.id_skip and all(
            s == 1 for s in block_args.strides
    ) and block_args.input_filters == block_args.output_filters:
        if drop_rate and (drop_rate > 0):
            x_in = Dropout(drop_rate,
                           noise_shape=(None, 1, 1, 1),
                           name=prefix + 'drop')(x_in)
        x_in = LAYERS.add([x_in, inputs], name=prefix + 'add')

    return x_in


def efficientnet_base(width_coefficient: float,  # pylint: disable=too-many-arguments, too-many-locals, too-many-branches
                      depth_coefficient: float,
                      default_resolution: int,
                      drop_connect_rate: float = 0.2,
                      depth_divisor: int = 8,
                      blocks_args=DEFAULT_BLOCKS_ARGS,
                      include_top: bool = True,
                      weights: str = 'imagenet',
                      input_tensor: Union[None, tf.Tensor] = None,
                      input_shape: Union[None, Tuple[int, int, int]] = None,
                      classes: int = 1000,
                      **kwargs: Any) -> List[tf.Tensor]:
    """Instantiates the EfficientNet architecture using given scaling coefficients.
    Optionally loads weights pre-trained on ImageNet.
    Note that the data format convention used by the model is
    the one specified in your Keras config at `~/.keras/keras.json`.
    # Arguments
        width_coefficient: float, scaling coefficient for network width.
        depth_coefficient: float, scaling coefficient for network depth.
        default_resolution: int, default input image size.
        dropout_rate: float, dropout rate before final classifier layer.
        drop_connect_rate: float, dropout rate at skip connections.
        depth_divisor: int.
        blocks_args: A tuple of BlockArgs to construct block modules.
        model_name: string, model name.
        include_top: whether to include the fully-connected
            layer at the top of the network.
        weights: one of `None` (random initialization),
              'imagenet' (pre-training on ImageNet),
              or the path to the weights file to be loaded.
        input_tensor: optional Keras tensor
            (i.e. output of `layers.Input()`)
            to use as image input for the model.
        input_shape: optional shape tuple, only to be specified
            if `include_top` is False.
            It should have exactly 3 inputs channels.
        pooling: optional pooling mode for feature extraction
            when `include_top` is `False`.
            - `None` means that the output of the model will be
                the 4D tensor output of the
                last convolutional layer.
            - `avg` means that global average pooling
                will be applied to the output of the
                last convolutional layer, and thus
                the output of the model will be a 2D tensor.
            - `max` means that global max pooling will
                be applied.
        classes: optional number of classes to classify images
            into, only to be specified if `include_top` is True, and
            if no `weights` argument is specified.
    # Returns
        A Keras model instance.
    # Raises
        ValueError: in case of invalid argument for `weights`,
            or invalid input shape.
    """
    global BACKEND, LAYERS, MODELS, KERAS_UTILS  # pylint: disable=global-statement
    BACKEND, LAYERS, MODELS, KERAS_UTILS = get_submodules_from_kwargs(kwargs)
    features = []
    if not (weights in {'imagenet', None} or os.path.exists(weights)):
        raise ValueError('The `weights` argument should be either '
                         '`None` (random initialization), `imagenet` '
                         '(pre-training on ImageNet), '
                         'or the path to the weights file to be loaded.')

    if weights == 'imagenet' and include_top and classes != 1000:
        raise ValueError('If using `weights` as `"imagenet"` with `include_top`'
                         ' as true, `classes` should be 1000')

    # Determine proper input shape
    input_shape = obtain_input_shape(input_shape,
                                     default_size=default_resolution,
                                     min_size=32,
                                     data_format=BACKEND.image_data_format(),
                                     require_flatten=include_top,
                                     weights=weights)

    if input_tensor is None:
        img_input = LAYERS.Input(shape=input_shape)
    else:
        if BACKEND.backend() == 'tensorflow':
            from tensorflow.python.keras.backend import is_keras_tensor  # pylint: disable=import-outside-toplevel
        else:
            is_keras_tensor = BACKEND.is_keras_tensor
        if not is_keras_tensor(input_tensor):
            img_input = LAYERS.Input(tensor=input_tensor, shape=input_shape)
        else:
            img_input = input_tensor

    bn_axis = 3 if BACKEND.image_data_format() == 'channels_last' else 1
    activation = get_swish(**kwargs)

    # Build stem
    x_in = img_input
    x_in = LAYERS.Conv2D(round_filters(32, width_coefficient, depth_divisor), 3,
                         strides=(2, 2),
                         padding='same',
                         use_bias=False,
                         kernel_initializer=CONV_KERNEL_INITIALIZER,
                         name='stem_conv')(x_in)
    x_in = LAYERS.BatchNormalization(axis=bn_axis, name='stem_bn')(x_in)
    x_in = LAYERS.Activation(activation, name='stem_activation')(x_in)
    # Build blocks
    num_blocks_total = sum(block_args.num_repeat for block_args in blocks_args)
    block_num = 0
    for idx, block_args in enumerate(blocks_args):
        assert block_args.num_repeat > 0
        # Update block input and output filters based on depth multiplier.
        block_args = block_args._replace(
            input_filters=round_filters(block_args.input_filters,
                                        width_coefficient, depth_divisor),
            output_filters=round_filters(block_args.output_filters,
                                         width_coefficient, depth_divisor),
            num_repeat=round_repeats(block_args.num_repeat, depth_coefficient))

        # The first block needs to take care of stride and filter size increase.
        drop_rate = drop_connect_rate * float(block_num) / num_blocks_total
        x_in = mb_conv_block(x_in, block_args,
                             activation=activation,
                             drop_rate=drop_rate,
                             prefix='block{}a_'.format(idx + 1))
        block_num += 1
        if block_args.num_repeat > 1:
            # pylint: disable=protected-access
            block_args = block_args._replace(
                input_filters=block_args.output_filters, strides=[1, 1])
            # pylint: enable=protected-access
            for bidx in xrange(block_args.num_repeat - 1):
                drop_rate = drop_connect_rate * float(block_num) / num_blocks_total
                block_prefix = 'block{}{}_'.format(
                    idx + 1,
                    string.ascii_lowercase[bidx + 1]
                )
                x_in = mb_conv_block(x_in, block_args,
                                     activation=activation,
                                     drop_rate=drop_rate,
                                     prefix=block_prefix)
                block_num += 1
        if idx < len(blocks_args) - 1 and blocks_args[idx + 1].strides[0] == 2:
            features.append(x_in)
        elif idx == len(blocks_args) - 1:
            features.append(x_in)
    return features


def efficientnet_b0(include_top: bool = True,
                    weights: str = 'imagenet',
                    input_tensor: Union[None, tf.Tensor] = None,
                    input_shape: Union[None, Tuple[int, int, int]] = None,
                    classes: int = 1000,
                    **kwargs: Any) -> List[tf.Tensor]:
    """EfficientNet-B0 model
    """
    return efficientnet_base(1.0, 1.0, 224, 0.2,
                             include_top=include_top, weights=weights,
                             input_tensor=input_tensor, input_shape=input_shape,
                             classes=classes,
                             **kwargs)


def efficientnet_b1(include_top: bool = True,
                    weights: str = 'imagenet',
                    input_tensor: Union[None, tf.Tensor] = None,
                    input_shape: Union[None, Tuple[int, int, int]] = None,
                    classes: int = 1000,
                    **kwargs: Any) -> List[tf.Tensor]:
    """EfficientNet-B1 model
    """
    return efficientnet_base(1.0, 1.1, 240, 0.2,
                             include_top=include_top, weights=weights,
                             input_tensor=input_tensor, input_shape=input_shape,
                             classes=classes, **kwargs)


def efficientnet_b2(include_top: bool = True,
                    weights: str = 'imagenet',
                    input_tensor: Union[None, tf.Tensor] = None,
                    input_shape: Union[None, Tuple[int, int, int]] = None,
                    classes: int = 1000,
                    **kwargs: Any) -> List[tf.Tensor]:
    """EfficientNet-B2 model
    """
    return efficientnet_base(1.1, 1.2, 260, 0.3,
                             include_top=include_top, weights=weights,
                             input_tensor=input_tensor, input_shape=input_shape,
                             classes=classes,
                             **kwargs)


def efficientnet_b3(include_top: bool = True,
                    weights: str = 'imagenet',
                    input_tensor: Union[None, tf.Tensor] = None,
                    input_shape: Union[None, Tuple[int, int, int]] = None,
                    classes: int = 1000,
                    **kwargs: Any) -> List[tf.Tensor]:
    """EfficientNet-B3 model
    """
    return efficientnet_base(1.2, 1.4, 300, 0.3,
                             include_top=include_top, weights=weights,
                             input_tensor=input_tensor, input_shape=input_shape,
                             classes=classes,
                             **kwargs)


def efficientnet_b4(include_top: bool = True,
                    weights: str = 'imagenet',
                    input_tensor: Union[None, tf.Tensor] = None,
                    input_shape: Union[None, Tuple[int, int, int]] = None,
                    classes: int = 1000,
                    **kwargs: Any) -> List[tf.Tensor]:
    """EfficientNet-B4 model
    """
    return efficientnet_base(1.4, 1.8, 380, 0.4,
                             include_top=include_top, weights=weights,
                             input_tensor=input_tensor, input_shape=input_shape,
                             classes=classes,
                             **kwargs)


def efficientnet_b5(include_top: bool = True,
                    weights: str = 'imagenet',
                    input_tensor: Union[None, tf.Tensor] = None,
                    input_shape: Union[None, Tuple[int, int, int]] = None,
                    classes: int = 1000,
                    **kwargs: Any) -> List[tf.Tensor]:
    """EfficientNet-B5 model
    """
    return efficientnet_base(1.6, 2.2, 456, 0.4,
                             include_top=include_top, weights=weights,
                             input_tensor=input_tensor, input_shape=input_shape,
                             classes=classes,
                             **kwargs)


def efficientnet_b6(include_top: bool = True,
                    weights: str = 'imagenet',
                    input_tensor: Union[None, tf.Tensor] = None,
                    input_shape: Union[None, Tuple[int, int, int]] = None,
                    classes: int = 1000,
                    **kwargs: Any) -> List[tf.Tensor]:
    """EfficientNet-B6 model
    """
    return efficientnet_base(1.8, 2.6, 528, 0.5,
                             include_top=include_top, weights=weights,
                             input_tensor=input_tensor, input_shape=input_shape,
                             classes=classes,
                             **kwargs)


def efficientnet_b7(include_top: bool = True,
                    weights: str = 'imagenet',
                    input_tensor: Union[None, tf.Tensor] = None,
                    input_shape: Union[None, Tuple[int, int, int]] = None,
                    classes: int = 1000,
                    **kwargs: Any) -> List[tf.Tensor]:
    """EfficientNet-B7 model
    """
    return efficientnet_base(2.0, 3.1, 600, 0.5,
                             include_top=include_top, weights=weights,
                             input_tensor=input_tensor, input_shape=input_shape,
                             classes=classes,
                             **kwargs)


setattr(efficientnet_b0, '__doc__', efficientnet_base.__doc__)
setattr(efficientnet_b1, '__doc__', efficientnet_base.__doc__)
setattr(efficientnet_b2, '__doc__', efficientnet_base.__doc__)
setattr(efficientnet_b3, '__doc__', efficientnet_base.__doc__)
setattr(efficientnet_b4, '__doc__', efficientnet_base.__doc__)
setattr(efficientnet_b5, '__doc__', efficientnet_base.__doc__)
setattr(efficientnet_b6, '__doc__', efficientnet_base.__doc__)
setattr(efficientnet_b7, '__doc__', efficientnet_base.__doc__)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
CLI functions for Peekingduck
"""

import os
import logging
import yaml
import click

from peekingduck.runner import Runner
from peekingduck.utils.logger import setup_logger
from peekingduck import __version__

setup_logger()
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


@click.group()
@click.version_option(__version__)
def cli() -> None:
    """
    PeekingDuck is a modular computer vision inference framework.

    Developed by Computer Vision Hub at AI Singapore.
    """


def _get_cwd() -> str:
    return os.getcwd()


def create_custom_folder(custom_folder_name: str) -> None:
    """Make custom nodes folder to create custom nodes
    """
    curdir = _get_cwd()
    custom_folder_dir = os.path.join(curdir, "src", custom_folder_name)
    custom_configs_dir = os.path.join(custom_folder_dir, 'configs')

    logger.info("Creating custom nodes folder in %s", custom_folder_dir)
    os.makedirs(custom_folder_dir, exist_ok=True)
    os.makedirs(custom_configs_dir, exist_ok=True)


def create_yml() -> None:
    """Inits the declarative yaml"""
    # Default yml to be discussed
    default_yml = dict(
        nodes=[
            'input.live',
            'model.yolo',
            'draw.bbox',
            'output.screen'
        ]
    )

    with open('run_config.yml', 'w') as yml_file:
        yaml.dump(default_yml, yml_file, default_flow_style=False)


@cli.command()
@click.option('--custom_folder_name', default='custom_nodes')
def init(custom_folder_name: str) -> None:
    """Initialise a PeekingDuck project"""
    print("Welcome to PeekingDuck!")
    create_custom_folder(custom_folder_name)
    create_yml()


@cli.command()
@click.option('--config_path', default=None, type=click.Path(),
              help="List of nodes to run. None assumes \
                   run_config.yml at current working directory")
@click.option('--node_config', default="None",
              help="""Modify node configs by wrapping desired configs in a JSON string.\n
                    Example: --node_config '{"node_name": {"param_1": var_1}}' """)
def run(config_path: str, node_config: str) -> None:
    """Runs PeekingDuck"""

    curdir = _get_cwd()
    if not config_path:
        config_path = os.path.join(curdir, "run_config.yml")

    runner = Runner(config_path, node_config, "src")
    runner.run()



# Copyright 2018 Ross Wightman
# Modifications copyright 2021 AI Singapore

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#      https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Supporting functions to decode multiple poses
"""


from typing import List, Tuple

import numpy as np
import tensorflow as tf
import scipy.ndimage as ndi
from peekingduck.pipeline.nodes.model.posenetv1.posenet_files.decode import decode_pose
from peekingduck.pipeline.nodes.model.posenetv1.posenet_files.constants import \
    LOCAL_MAXIMUM_RADIUS, SWAP_AXES


def decode_multiple_poses(model_output: Tuple[np.ndarray, tf.Tensor, tf.Tensor, tf.Tensor],
                          dst_keypoint_scores: np.ndarray,
                          dst_keypoints: np.ndarray,
                          output_stride: int,
                          score_threshold: float = 0.5,
                          nms_radius: int = 20,
                          min_pose_score: float = 0.5) -> int:
    # pylint: disable=too-many-arguments
    """ Decodes the offsets and displacements map and return the keypoints


    Args:
        model_output (Tuple[np.ndarray, tf.Tensor, tf.Tensor, tf.Tensor]): scores,
                offsets, displacements_fwd, displacements_bwd from model predictions
        dst_scores (np.array): Nx17 buffer to store keypoint scores where N is
                        the max persons to be detected
        dst_keypoints (np.array): Nx17x2 buffer to store keypoints coordinate
                        where N is the max persons to be detected
        output_stride (int): output stride to convert output indices to image coordinates
        score_threshold (float): return instance detections if root part score
                        >= threshold
        nms_radius (int): non-maximum suppression part distance in pixels
        min_pose_score (float): minimum pose score to return detected pose

    Returns:
        pose_count (int): number of poses detected
    """
    scores, offsets, displacements_fwd, displacements_bwd = model_output
    scores = np.array(scores[0])
    offsets = np.array(offsets[0])
    displacements_fwd = np.array(displacements_fwd[0])
    displacements_bwd = np.array(displacements_bwd[0])

    scored_parts = _build_part_with_score_fast(score_threshold,
                                               LOCAL_MAXIMUM_RADIUS,
                                               scores)
    scored_parts = _sort_scored_parts(scored_parts)

    offsets, displacements_fwd, displacements_bwd = _change_dimensions(
        scores,
        offsets,
        displacements_fwd,
        displacements_bwd)

    pose_count = _look_for_poses(
        scored_parts,
        scores,
        offsets,
        displacements_fwd,
        displacements_bwd,
        dst_keypoint_scores,
        dst_keypoints,
        output_stride,
        nms_radius,
        min_pose_score)

    return pose_count


def _build_part_with_score_fast(
        score_threshold: float,
        local_max_radius: int,
        scores: np.ndarray) -> List[Tuple[float, int, np.ndarray]]:
    """ Returns an array of parts with score, id and coordinate in each part """
    parts = []
    lmd = 2 * local_max_radius + 1

    max_vals = ndi.maximum_filter(scores, size=(lmd, lmd, 1), mode='constant')
    max_loc = np.logical_and(
        scores == max_vals, scores > score_threshold)
    max_loc_idx = max_loc.nonzero()
    for y_coord, x_coord, keypoint_id in zip(*max_loc_idx):
        parts.append((scores[y_coord, x_coord, keypoint_id],
                      keypoint_id,
                      np.array((x_coord, y_coord))))

    return parts


def _sort_scored_parts(
        parts: List[Tuple[float, int, np.ndarray]]) -> List[Tuple[float, int, np.ndarray]]:
    """ Sort parts by confidence scores """
    parts = sorted(parts, key=lambda x: x[0], reverse=True)
    return parts


def _change_dimensions(scores: np.ndarray,
                       offsets: np.ndarray,
                       displacements_fwd: np.ndarray,
                       displacements_bwd: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """ Change dimensions from (h, w, x) to (h, w, x//2, 2) to return a
    complete coordinate array
    """
    height = scores.shape[0]
    width = scores.shape[1]
    offsets = offsets.reshape(height, width, 2, -1).swapaxes(2, 3)
    displacements_fwd = displacements_fwd.reshape(
        height, width, 2, -1).swapaxes(2, 3)
    displacements_bwd = displacements_bwd.reshape(
        height, width, 2, -1).swapaxes(2, 3)

    offsets = offsets[:, :, :, SWAP_AXES]
    displacements_fwd = displacements_fwd[:, :, :, SWAP_AXES]
    displacements_bwd = displacements_bwd[:, :, :, SWAP_AXES]

    return offsets, displacements_fwd, displacements_bwd


def _look_for_poses(scored_parts: List[Tuple[float, int, np.ndarray]],
                    scores: np.ndarray,
                    offsets: np.ndarray,
                    displacements_fwd: np.ndarray,
                    displacements_bwd: np.ndarray,
                    dst_keypoint_scores: np.ndarray,
                    dst_keypoints: np.ndarray,
                    output_stride: int,
                    nms_radius: int,
                    min_pose_score: float) -> int:
    # pylint: disable=too-many-arguments, too-many-locals
    """ Change dimensions from (h, w, x) to (h, w, x//2, 2) to return a
    complete coordinate array
    """
    pose_count = 0
    dst_keypoint_scores[:] = 0
    max_pose_detections = dst_keypoint_scores.shape[0]
    squared_nms_radius = nms_radius**2

    for root_score, root_id, root_coord in scored_parts:
        root_image_coords = _calculate_keypoint_coords_on_image(root_coord,
                                                                output_stride,
                                                                offsets,
                                                                root_id)

        if _within_nms_radius_fast(dst_keypoints[:pose_count, root_id, :],
                                   squared_nms_radius,
                                   root_image_coords):
            continue

        keypoint_scores = dst_keypoint_scores[pose_count]
        keypoint_coords = dst_keypoints[pose_count]
        decode_pose(root_score,
                    root_id,
                    root_image_coords,
                    scores,
                    offsets,
                    output_stride,
                    displacements_fwd,
                    displacements_bwd,
                    keypoint_scores,
                    keypoint_coords)

        pose_score = _get_instance_score_fast(dst_keypoints[:pose_count, :, :],
                                              squared_nms_radius,
                                              keypoint_scores,
                                              keypoint_coords)
        if min_pose_score == 0. or pose_score >= min_pose_score:
            pose_count += 1
        else:
            dst_keypoint_scores[pose_count] = 0

        if pose_count >= max_pose_detections:
            break

    return pose_count


def _calculate_keypoint_coords_on_image(heatmap_positions: np.ndarray,
                                        output_stride: int,
                                        offsets: np.ndarray,
                                        keypoint_id: int) -> np.ndarray:
    """ Calculate keypoint image coordinates from heatmap positions,
        output_stride and offset_vectors
    """
    offset_vectors = offsets[heatmap_positions[1],
                             heatmap_positions[0],
                             keypoint_id]
    return heatmap_positions * output_stride + offset_vectors


def _within_nms_radius_fast(pose_coords: np.ndarray,
                            squared_nms_radius: int,
                            point: np.ndarray) -> bool:
    """ Check if keypoint is within squared nms radius
    """
    if not pose_coords.shape[0]:
        return False
    return np.any(
        np.sum((pose_coords - point)**2, axis=1) <= squared_nms_radius)


def _get_instance_score_fast(exist_pose_coords: np.ndarray,
                             squared_nms_radius: int,
                             keypoint_scores: np.ndarray,
                             keypoint_coords: np.ndarray) -> float:
    """ Obtain instance scores for keypoints
    """
    if exist_pose_coords.shape[0]:
        score_sum = np.sum((exist_pose_coords - keypoint_coords) ** 2, axis=2)\
            > squared_nms_radius
        not_overlapped_scores = np.sum(keypoint_scores[np.all(score_sum, axis=0)])
    else:
        not_overlapped_scores = np.sum(keypoint_scores)
    return not_overlapped_scores / len(keypoint_scores)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Reads a videofeed from a stream (e.g. webcam)
"""

from typing import Dict, Any

from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.input.utils.preprocess import resize_image
from peekingduck.pipeline.nodes.input.utils.read import VideoThread, VideoNoThread


class Node(AbstractNode):
    """Node to receive livestream as inputs.

    Inputs:
        None

    Outputs:
        |img|

        |filename|

        |pipeline_end|

        |saved_video_fps|

    Configs:
        fps_saved_output_video (:obj:`int`): **default = 10**

            FPS of the mp4 file after livestream is processed and exported.
            FPS dependent on running machine performance.

        filename (:obj:`str`):  **default = "webcamfeed.mp4"**

            Filename of the mp4 file if livestream is exported.

        resize (:obj:`Dict`): **default = { do_resizing: False, width: 1280, height: 720 }**

            Dimension of extracted image frame

        input_source (:obj:`int`): **default = 0 (for webcam)**

            Refer to `OpenCV doucmentation
            <https://docs.opencv.org/3.4/d4/d15/group__videoio__flags__base.html#ga023786be1ee68a9105bf2e48c700294d/>`_
            for list of source stream codes

        mirror_image (:obj:`bool`): **default = False**

            Boolean to set extracted image frame as mirror image of input stream

        frames_log_freq (:obj:`int`): **default = 100**

            Logs frequency of frames passed in cli

        threading (:obj:`bool`): **default = False**

            Boolean to enable threading when reading frames from camera.
            The FPS can increase up to 30% if this is enabled for Windows and MacOS.
            This will also be supported in Linux in future PeekingDuck versions.

    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self._allowed_extensions = ["mp4", "avi", "mov", "mkv"]
        if self.threading:
            self.videocap = VideoThread(                # type: ignore
                self.input_source, self.mirror_image)
        else:
            self.videocap = VideoNoThread(              # type: ignore
                self.input_source, self.mirror_image)

        width, height = self.videocap.resolution
        self.logger.info('Device resolution used: %s by %s', width, height)
        if self.resize['do_resizing']:
            self.logger.info('Resizing of input set to %s by %s',
                             self.resize['width'],
                             self.resize['height'])
        if self.filename.split('.')[-1] not in self._allowed_extensions:
            raise ValueError("filename extension must be one of: ",
                             self._allowed_extensions)

        self.frame_counter = 0

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        success, img = self.videocap.read_frame()  # type: ignore

        if success:
            if self.resize['do_resizing']:
                img = resize_image(img,
                                   self.resize['width'],
                                   self.resize['height'])

            outputs = {"img": img,
                       "pipeline_end": False,
                       "filename": self.filename,
                       "saved_video_fps": self.fps_saved_output_video}
            self.frame_counter += 1
            if self.frame_counter % self.frames_log_freq == 0:
                self.logger.info('Frames Processed: %s ...',
                                 self.frame_counter)

        else:
            outputs = {"img": None,
                       "pipeline_end": True,
                       "filename": self.filename,
                       "saved_video_fps": self.fps_saved_output_video}
            self.logger.warning("No video frames available for processing.")

        return outputs



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Checks for model weights in weights folder
"""


import os
from typing import List


def has_weights(root: str, path_to_check: List[str]) -> bool:
    """Checks for model weight paths from weights folder

    Args:
        root (str): path of peekingduck root folder
        path_to_check (List[str]): list of files/directories to check
            to see if weights exists

    Returns:
        boolean: True is files/directories needed exists, else False
    """

    # Check for whether weights dir even exist. If not make directory
    # Empty directory should then return False
    weights_dir = os.path.join(root, '..', 'weights')
    if not os.path.isdir(weights_dir):
        os.mkdir(weights_dir)
        return False

    for check in path_to_check:
        if not os.path.exists(os.path.join(root, check)):
            return False
    return True



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
functions for drawing bounding box related UI components
"""

from typing import List, Tuple
import numpy as np
import cv2
from cv2 import FONT_HERSHEY_SIMPLEX, LINE_AA
from peekingduck.pipeline.nodes.draw.utils.constants import \
    CHAMPAGNE, BLACK, THICK, VERY_THICK, NORMAL_FONTSCALE, \
    POINT_RADIUS, FILLED, PRIMARY_PALETTE, PRIMARY_PALETTE_LENGTH as TOTAL_COLOURS
from peekingduck.pipeline.nodes.draw.utils.general import \
    get_image_size, project_points_onto_original_image


def draw_bboxes(frame: np.array,
                bboxes: List[List[float]],
                bbox_labels: List[str],
                show_labels: bool) -> None:
    """Draw bboxes onto an image frame.

    Args:
        frame (np.array): image of current frame
        bboxes (List[List[float]]): bounding box coordinates
        colour (Tuple[int, int, int]): colour used for bounding box
        bbox_labels (List[str]): labels of object detected
    """
    image_size = get_image_size(frame)
    # Get unique label colour indexes
    colour_indx = {label: indx for indx, label in enumerate(set(bbox_labels))}

    for i, bbox in enumerate(bboxes):
        colour = PRIMARY_PALETTE[colour_indx[bbox_labels[i]] % TOTAL_COLOURS]
        if show_labels:
            _draw_bbox(frame, bbox, image_size,
                       colour, bbox_labels[i])
        else:
            _draw_bbox(frame, bbox, image_size, colour)

def _draw_bbox(frame: np.array,
               bbox: List[float],
               image_size: Tuple[int, int],
               colour: Tuple[int, int, int],
               bbox_label: str = None) -> np.array:
    """ Draw a single bounding box """
    top_left, bottom_right = project_points_onto_original_image(
        bbox, image_size)
    cv2.rectangle(frame, (top_left[0], top_left[1]),
                  (bottom_right[0], bottom_right[1]),
                  colour, VERY_THICK)

    if bbox_label:
        _draw_label(frame, top_left, bbox_label, colour, BLACK)


def _draw_label(frame: np.array,
                top_left: Tuple[int, int],
                bbox_label: str,
                bg_colour: Tuple[int, int, int],
                text_colour: Tuple[int, int, int]) -> None:
    """Draw bbox label at top left of bbox"""
        # get label size
    (text_width, text_height), baseline = cv2.getTextSize(bbox_label,
                                                          FONT_HERSHEY_SIMPLEX,
                                                          NORMAL_FONTSCALE,
                                                          THICK)
    # put filled text rectangle
    cv2.rectangle(frame,
                  (top_left[0], top_left[1]),
                  (int(top_left[0] + text_width),
                   int(top_left[1] - text_height - baseline)),
                  bg_colour,
                  FILLED)

    # put text above rectangle
    bbox_label = bbox_label[:1].capitalize() + bbox_label[1:]
    cv2.putText(frame, bbox_label, (top_left[0], int(top_left[1]-6)), FONT_HERSHEY_SIMPLEX,
                NORMAL_FONTSCALE, text_colour, THICK, LINE_AA)


def draw_tags(frame: np.array,
              bboxes: List[List[float]],
              tags: List[str],
              colour: Tuple[int, int, int]) -> None:
    """Draw tags above bboxes.

    Args:
        frame (np.array): image of current frame
        bboxes (List[List[float]]): bounding box coordinates
        tags (List[string]): tag associated with bounding box
        color (Tuple[int, int, int]): color of text
    """
    image_size = get_image_size(frame)
    for idx, bbox in enumerate(bboxes):
        _draw_tag(frame, bbox, tags[idx], image_size, colour)


def _draw_tag(frame: np.array,
              bbox: np.array,
              tag: str,
              image_size: Tuple[int, int],
              colour: Tuple[int, int, int]) -> None:
    """Draw a tag above a single bounding box.
    """
    top_left, btm_right = project_points_onto_original_image(bbox, image_size)

    # Find offset to centralize text
    (text_width, _), baseline = cv2.getTextSize(tag,
                                                FONT_HERSHEY_SIMPLEX,
                                                NORMAL_FONTSCALE,
                                                THICK)
    bbox_width = btm_right[0] - top_left[0]
    offset = int((bbox_width - text_width) / 2)
    position = (int(top_left[0] + offset), int(top_left[1] - baseline))
    cv2.putText(frame, tag, position, FONT_HERSHEY_SIMPLEX,
                NORMAL_FONTSCALE, colour, VERY_THICK)



def draw_pts(frame: np.array, pts: List[Tuple[float]]) -> None:
    """draw pts of selected object onto frame

    Args:
        frame (np.array): image of current frame
        pts (List[Tuple[float]]): bottom midpoints of bboxes
    """
    for point in pts:
        cv2.circle(frame, point, POINT_RADIUS, CHAMPAGNE, -1)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Fast Object Detection model
"""

from typing import Dict, Any
from peekingduck.pipeline.nodes.node import AbstractNode
from .yolov4 import yolo_model


class Node(AbstractNode):
    """Yolo node class that initialises and use yolo model to infer bboxes
    from image frame.

    The yolo node is capable of detecting objects from 80 categories. It uses
    YOLOv4-tiny by default and can be changed to using YOLOv4. The table of
    categories can be found :term:`here <object detection indices>`.

    Inputs:
        |img|

    Outputs:
        |bboxes|

        |bbox_labels|

        |bbox_scores|

    Configs:
        model_type (:obj:`str`): **{"v4", "v4tiny"}, default="v4tiny"**

            defines the type of YOLO model to be used.

        weights_dir (:obj:`List`):
            directory pointing to the model weights.

        blob_file (:obj:`str`):
            name of file to be downloaded, if weights are not found in `weights_dir`.

        graph_files (:obj:`Dict`):
            dictionary pointing to path of the model weights file.

        size (:obj:`int`): **default = 416 **

            image resolution passed to the YOLO model.

        num_classes (:obj:`int`): **default = 80 **

            maximum number of objects to be detected.

        detect_ids (:obj:`List`): **default = [0] **

            list of object class ids to be detected.

        max_output_size_per_class (:obj:`int`): **default = 50 **

            maximum number of detected instances for each class in an image.

        max_total_size (:obj:`int`): **default = 50 **

            maximum total number of detected instances in an image.

        yolo_iou_threshold (:obj:`float`): **[0,1], default = 0.5**

            overlapping bounding boxes above the specified IoU (Intersection
            over Union) threshold are discarded.

        yolo_score_threshold (:obj:`float`): **[0,1], default = 0.2**

            bounding box with confidence score less than the specified
            confidence score threshold is discarded.

    References:

    YOLOv4: Optimal Speed and Accuracy of Object Detection:
        https://arxiv.org/pdf/2004.10934v1.pdf

    Model weights trained by https://github.com/hunglc007/tensorflow-yolov4-tflite
    Inference code adapted from https://github.com/zzh8829/yolov3-tf2
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.model = yolo_model.YoloModel(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """function that reads the image input and returns the bboxes
        of the specified objects chosen to be detected

        Args:
            inputs (Dict): Dictionary of inputs with key "img"

        Returns:
            outputs (Dict): bbox output in dictionary format with keys
            "bboxes", "bbox_labels" and "bbox_scores"
        """
        # Currently prototyped to return just the bounding boxes
        # without the scores
        bboxes, labels, scores = self.model.predict(inputs["img"])
        outputs = {"bboxes": bboxes, "bbox_labels": labels, "bbox_scores": scores}
        return outputs



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Yolo model with model types: v3 and v3tiny
"""

import os
import logging
from typing import List, Dict, Any, Tuple
import numpy as np
from peekingduck.weights_utils import checker, downloader
from .yolo_files.detector import Detector


class YoloModel:
    """Yolo model with model types: v4 and v4tiny"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__()

        self.logger = logging.getLogger(__name__)

        # check threshold values
        if not 0 <= config['yolo_score_threshold'] <= 1:
            raise ValueError("yolo_score_threshold must be in [0, 1]")

        # check for yolo weights, if none then download into weights folder
        if not checker.has_weights(config['root'],
                                   config['weights_dir']):
            self.logger.info('---no yolo weights detected. proceeding to download...---')
            downloader.download_weights(config['root'],
                                        config['blob_file'])
            self.logger.info('---yolo weights download complete.---')

        # get classnames path to read all the classes
        classes_path = os.path.join(config['root'], config['classes'])
        self.class_names = [c.strip() for c in open(classes_path).readlines()]
        self.detect_ids = config['detect_ids']

        self.detector = Detector(config)

    def predict(self, frame: np.array) -> Tuple[List[np.array], List[str], List[float]]:
        """predict the bbox from frame

        returns:
        object_bboxes(List[Numpy Array]): list of bboxes detected
        object_labels(List[str]): list of string labels of the
            object detected for the corresponding bbox
        object_scores(List(float)): list of confidence scores of the
            object detected for the corresponding bbox
        """
        assert isinstance(frame, np.ndarray)

        # return bboxes, object_bboxes, object_labels, object_scores
        return self.detector.predict_object_bbox_from_image(
            self.class_names, frame, self.detect_ids
        )

    def get_detect_ids(self) -> List[int]:
        """getter for selected ids for detection

        Returns:
            List[int]: list of selected detection ids
        """
        return self.detect_ids



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Assign objects in close proximity to groups
"""

from typing import Any, Dict, List, Tuple

import numpy as np

from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.dabble.utils.quick_find import QuickFind


class Node(AbstractNode):
    """Node that groups objects that are near each other.

    It does so by comparing the 3D location of all objects, and assigning objects
    near each other to the same group.

    Inputs:
        |obj_3D_locs|

    Outputs:
        |obj_groups|

    Configs:
        obj_dist_thres (:obj:`float`): **default = 1.5**

            Threshold of distance, in metres, between two objects, less than which they
            would be assigned to the same group.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """ Checks the distance between 3d locations of a pair of objects.
        If distance is less than threshold, assign the objects to the same group.
        Repeat for all object pairs.
        """

        nearby_obj_pairs = self._find_nearby_obj_pairs(
            inputs["obj_3D_locs"], self.obj_dist_thres)

        quickfind = QuickFind(len(inputs["obj_3D_locs"]))
        for (idx_1, idx_2) in nearby_obj_pairs:
            if not quickfind.connected(idx_1, idx_2):
                quickfind.union(idx_1, idx_2)

        return {"obj_groups": quickfind.get_group_alloc()}

    @staticmethod
    def _find_nearby_obj_pairs(obj_locs: List[np.array],
                               obj_dist_thres: float) -> List[Tuple[int, int]]:
        """If the distance between 2 objects are less than the threshold, append their
        indexes to nearby_obj_pairs as a tuple."""

        nearby_obj_pairs = []
        for idx_1, loc_1 in enumerate(obj_locs):
            for idx_2, loc_2 in enumerate(obj_locs):
                if idx_1 == idx_2:
                    continue

                dist_bet = np.linalg.norm(loc_1 - loc_2)

                if dist_bet <= obj_dist_thres:
                    if (idx_2, idx_1) not in nearby_obj_pairs:
                        nearby_obj_pairs.append((idx_1, idx_2))

        return nearby_obj_pairs



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Make trainable=False freeze BN
"""

import numpy as np
import tensorflow as tf


class BatchNormalization(tf.keras.layers.BatchNormalization):  # pylint: disable=too-many-ancestors, too-few-public-methods
    """
    Make trainable=False freeze BN for real (the og version is sad)
    """

    def call(self, inputs: np.array, training: tf.Tensor = None) -> object:
        """Make trainable=False freeze BN for real (the og version is sad)

        Input:
            - inputs: input image matrix
            - training: the training data

        Output:
            - inputs_image: transformed image matrix
        """
        if training is None:
            training = tf.constant(False)
        training = tf.logical_and(training, self.trainable)
        return super().call(inputs, training)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Main engine for Peekingduck processes
"""

import copy
import sys
import logging
from typing import List
from peekingduck.pipeline.pipeline import Pipeline
from peekingduck.declarative_loader import DeclarativeLoader
from peekingduck.pipeline.nodes.node import AbstractNode


class Runner():
    """
    The runner class for creation of pipeline using declared/given nodes.

    The runner class uses the provided configurations to setup a node pipeline
    which is used to run inference.

    Args:

        RUN_PATH (:obj:`str`): If path to a run_config.yml is provided, uses \
        our declarative loader to load the yaml file according to our specified \
        schema to obtain the declared nodes that would be sequentially \
        initialized and used to create the pipeline for running inference. \

        config_updates_cli (:obj:`str`): config changes passed as part of the \
        cli command sed to modify the node configurations direct from cli.

        CUSTOM_NODE_PARENT_FOLDER (:obj:`str`): path to folder which contains \
        custom nodes that users have created to be used with PeekingDuck. \
        For more information on using custom nodes, please refer to \
        `getting started <getting_started/03_custom_nodes.html>`_.

        nodes (:obj:`list` of :obj:`Node`): if not using declarations via yaml, \
        initialize by giving the node stack directly as a list.

    """

    def __init__(self,
                 RUN_PATH: str = "",
                 config_updates_cli: str = None,
                 CUSTOM_NODE_PARENT_FOLDER: str = None,
                 nodes: List[AbstractNode] = None):

        self.logger = logging.getLogger(__name__)

        if not nodes and RUN_PATH:
            # create Graph to run
            self.node_loader = DeclarativeLoader(
                RUN_PATH, config_updates_cli, CUSTOM_NODE_PARENT_FOLDER)  # type: ignore

            self.pipeline = self.node_loader.get_pipeline()

        # If Runner given nodes, instantiated_nodes is created differently
        else:
            try:
                self.pipeline = Pipeline(nodes) # type: ignore
            except ValueError as error:
                self.logger.error(str(error))
                sys.exit(1)

    def run(self) -> None:
        """execute single or continuous inference
        """
        while not self.pipeline.terminate:
            for node in self.pipeline.nodes:
                if "pipeline_end" in self.pipeline.data and \
                        self.pipeline.data["pipeline_end"]:  # type: ignore

                    self.pipeline.terminate = True
                    if "pipeline_end" not in node.inputs:
                        continue

                if "all" in node.inputs:
                    inputs = copy.deepcopy(self.pipeline.data)
                else:
                    inputs = {key: self.pipeline.data[key]
                              for key in node.inputs if key in self.pipeline.data}

                outputs = node.run(inputs)
                self.pipeline.data.update(outputs)  # type: ignore

    def get_run_config(self) -> List[str]:
        """retrieve run configs

        Returns:
            (:obj:`Dict`: run configs being used for runner
        """
        return self.node_loader.node_list
